'''
Timeseries.py - functions associated with analysis of time series data
======================================================================
:Author: Mike Morgan
:Release: $Id$
:Date: |today|
:Tags: Python

Functions are split into::
 * clustering assessment
 * data transformation and normalisation
 * differential expression analysis
 * clustering and distance metric functions
'''
import sklearn.metrics.cluster.supervised as supervised
from math import log
import CGAT.Experiment as E
import numpy as np
import pandas as pd
import itertools
import os
import math
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import importr
from rpy2.robjects import r as R
import rpy2.robjects as ro
import random
from CGAT.Timeseries import cmetrics as c2m


def get_r_path():
    """return path of R support functions.
    """
    return os.path.dirname(__file__)



def get_label_map(labels):
    '''
    return a dictionary with integer:string mapping
    '''
    label_set = set()
    map_dict = {}
    for val in labels:
        label_set.update(val)
    for lab, integer in enumerate(label_set):
        map_dict[integer] = lab

    return map_dict


def make_mapped_matrix(map_dict, input_frame):
    '''
    return a matrix with integer labels from mapping
    '''

    frame_index = input_frame.index.tolist()
    nindex = len(frame_index)
    ncols = len(input_frame.columns)
    integer_matrix = np.ndarray((nindex, ncols),
                                dtype=np.int32)

    E.info("mapping cluster labels")
    matrix_idx = [h for h, g in enumerate(frame_index)]
    for idx in matrix_idx:
        for col in range(ncols):
            mod = input_frame.iloc[idx][col+1]
            integer_matrix[idx][col] = map_dict[mod]

    return integer_matrix


def randIndexes(clustering_results):
    '''
    Calculate Rand index and adjusted Rand index over pairwise
    clustering comparisons.
    Use cythonised function to calculate indices
    '''

    # reassign module and gene labels with integer ids, integer comparison is
    # much faster than string comparison
    cluster_labels = clustering_results.values
    map_dict = get_label_map(cluster_labels)

    gene_map = {}
    for r, gene in enumerate(clustering_results.index):
        gene_map[gene] = r
    E.info("mapping gene ids")

    integer_matrix = make_mapped_matrix(map_dict, clustering_results)
    # take a small slice of the matrix for testing 5 genes, 3 clusterings

    E.info("counting clustering consensus")
    # use cythonized function to return rand index matrix
    cy_rand = c2m.consensus_metrics(integer_matrix)
    E.info("Rand Index calculated for all clusterings")

    return cy_rand


def unravel_arrays(metric_array):
    '''
    Unravel a numpy array such that only one half of the symmetrical
    matrix is output.  Do not output diagonal values.
    '''

    dim = metric_array.shape[0]
    flat_array = []

    for indx in itertools.combinations(range(0, dim), r=2):
        if indx[0] != indx[1]:
            flat_array.append(metric_array[indx[1], indx[0]])
        else:
            pass
    return flat_array


def mutualInformation(cluster1, cluster2):
    '''
    Calculate the mutual information for a given pair of clusterings.
    Assume clustering1 represents the reference ground truth.
    Code from scikit-learn.

    The mutual information of two clusterings, U and V is given by:

    MI(U, V) = sum(R)sum(C)P(i, j) [log(P(i, j)/(p(i)p'(j)))]

    This is the similarity of two clustering labels, where P(i) is
    the probability of a random sample in clustering Ui, and P'(j)
    the probability of a random sample in clustering Vj.
    '''

    cont = contingency(cluster1, cluster2)
    cont_sum = np.sum(cont)
    pi = np.sum(cont, axis=1)
    pj = np.sum(cont, axis=0)
    outer = np.outer(pi, pj)
    nnz = cont != 0

    cont_nm = cont[nnz]
    log_cont_nm = np.log(cont_nm)
    cont_nm /= cont_sum
    log_outer = -np.log(outer[nnz]) + log(pi.sum()) + log(pj.sum())
    mi = (cont_nm * (log_cont_nm - log(cont_sum)) + (cont_nm * log_outer))

    return mi.sum()


def contingency(cluster1, cluster2):
    '''
    Return a n x m matrix of clustering overlaps, where n
    is the number of clusters in clustering1 and m is the
    number of clusters in clustering2.  Return an np array.
    '''
    cont = pd.DataFrame(columns=cluster1.keys(), index=cluster2.keys())
    cont = cont.fillna(0.0)

    for x in itertools.product(cluster1.keys(), cluster2.keys()):
        set1 = cluster1[x[0]]
        set2 = cluster2[x[1]]
        intersect = len(set1.intersection(set2))
        cont[x[0]][x[1]] = intersect

    cont = cont.as_matrix()
    return cont


def entropy(cluster_labels):
    '''
    Calculate the entropy of a clustering.
    Entropy(X):
    H(X) = sum(p(i)log(1/(pi)))
    '''

    if len(cluster_labels) == 0:
        return 1.0
    else:
        pass

    cluster_prob = [len(cluster_labels[x]) for x in cluster_labels.keys()]
    pi = np.array(cluster_prob).astype(np.float)
    pi = pi[pi > 0]
    pi_sum = np.sum(pi)

    entropy = -np.sum((pi / pi_sum) * (np.log(pi) - log(pi_sum)))

    return entropy


def adjustedMutualInformation(cluster1, cluster2):
    '''
    Using the scikit-learn algorithms, calculate the adjusted mutual
    information for two clusterings.  Assume cluster1 is the
    reference/ground truth clustering.
    The adjusted MI accounts for higher scores by chance, particularly
    in the case where a larger number of clusters leads to a higher MI.

    AMI(U, V) = [MI(U, V) - E(MI(U, V))] / [max(H(U), H(V)) - E(MI(U, V))]
    where E(MI(U, V)) is the expected mutual information given the
    number of clusters and H(U), H(V) are the entropies of clusterings
    U and V.
    '''

    cont = contingency(cluster1, cluster2)
    mi = mutualInformation(cluster1, cluster2)
    sample_size = float(sum([len(cluster1[x]) for x in cluster1.keys()]))

    # Given the number of samples, what is the expected number
    # of overlaps that would occur by chance?
    emi = supervised.expected_mutual_information(cont, sample_size)

    # calculate the entropy for each clustering
    h_clust1, h_clust2 = entropy(cluster1), entropy(cluster2)

    if abs(h_clust1) == 0.0:
        h_clust1 = 0.0
    else:
        pass
    if abs(h_clust2) == 0.0:
        h_clust2 = 0.0
    else:
        pass

    ami = (mi - emi) / (max(h_clust1, h_clust2) - emi)

    # bug: entropy will return -0 in some instances
    # make sure this is actually 0 else ami will return None
    # instead of 0.0

    if np.isnan(ami):
        ami = np.nan_to_num(ami)
    else:
        pass

    return ami




def deseqNormalize(infile,
                   time_points,
                   reps,
                   conditions=None):
    '''
    Library size normalisation and variance stabilizing transformation of
    timeseries RNA-seq data

    :param infile: count table from NGS-seq experiment
    :type infile: str
    :param time_points: time point labels
    :type time_points: str list
    :param reps: replicates labels
    :type reps: str list
    :param conditions: if  multiple experimental conditions
    are to be normalised at the same time
    :type conditions: str list
    '''
    # MM: NB - this should be split into separate library size
    # normalisation and VST transformations
    # maybe add in different transformation options.

    pandas2ri.activate()
    reps = reps

    # load library
    R('''suppressMessages(library("DESeq"))''')

    # generates a lists for the design data frame
    # of the proper length
    # these need to be rpy2 objects to be parsed
    # properly in the string formatting

    E.info("converting to pandas dataframe object")

    if infile.split(".")[-1] == "gz":
        comp = "gzip"
    else:
        comp = None

    data_frame = pd.read_table(infile,
                               index_col=0,
                               header=0,
                               sep="\t",
                               compression=comp)
    # py2ri requires activation
    pandas2ri.activate()
    rdf = pandas2ri.py2ri(data_frame)

    if not conditions:
        time_rep_comb = [x for x in itertools.product(time_points, reps)]
        time_cond = ro.StrVector([x[0] for x in time_rep_comb])
        rep_cond = ro.StrVector([x[1] for x in time_rep_comb])

        R.assign('countsTable', rdf)
        R('''design <- data.frame(row.names=colnames(countsTable),'''
          '''times=%s, replicates=%s)''' % (time_cond.r_repr(),
                                            rep_cond.r_repr()))
    elif conditions:
        design_dict = {}
        for x in data_frame.columns.values:
            sample_dict = {}
            sample_dict['condition'] = str(x).split(".")[0]
            sample_dict['times'] = int(str(x).split(".")[1])
            sample_dict['replicates'] = str(x).split(".")[2]
            design_dict[x] = sample_dict
            design_frame = pd.DataFrame(design_dict)
            design_frame = design_frame.T

        des_cond = design_frame['condition'].values.tolist()
        des_time = design_frame['times'].values.tolist()
        des_reps = design_frame['replicates'].values.tolist()

        cond_cond = ro.StrVector([x for x in des_cond])
        time_cond = ro.StrVector([x for x in des_time])
        rep_cond = ro.StrVector([x for x in des_reps])

        R.assign('countsTable', rdf)
        R.assign('design', design_frame)

    # create the count data set and normalize to library size
    # transform with variance stabilizing transformation
    # only select genes with an average of ten reads mapping

    E.info("calculating size factors and dispersion")
    R('''notZero <- (rowMeans(countsTable) > 1)''')
    R('''cds <- newCountDataSet(countsTable[notZero, ], design)''')
    R('''cds_size <- estimateSizeFactors(cds)''')
    R('''cds_disp <- estimateDispersions(cds_size, method="blind")''')

    E.info("applying variance stabilizing transformation")

    R('''vst <- varianceStabilizingTransformation(cds_disp)''')

    # format data set to long format with condition and replicate labels
    # convert to a numpy array

    R('''replicates <- c(%s)''' % rep_cond.r_repr())
    R('''times <- c(%s)''' % time_cond.r_repr())
    if conditions:
        R('''conditions <- c(%s)''' % cond_cond.r_repr())
        R('''trans_vst = data.frame(t(exprs(vst)), '''
          '''times, replicates, conditions)''')
    else:
        R('''trans_vst = data.frame(t(exprs(vst)), times, replicates)''')

    # load data and convert to pandas object
    data_file = pandas2ri.ri2py(R["trans_vst"])
    
    return data_file


def avTimeExpression(infile):
    '''
    Calculate average expression over replicates at each time point
    Requires genes as columns with 'replicates' and 'times' as additional
    columns
    '''

    # check file compression
    if infile.split(".")[-1] == "gz":
        comp = "gzip"
    else:
        comp = None

    df = pd.read_table(infile, sep="\t",
                       header=0, index_col=0,
                       compression=comp)

    # average over replicates at each time point
    # then recombine
    df_groups = df.groupby(by='times')
    data_frame = pd.DataFrame(index=df.columns,
                              columns=None)
    for names, groups in df_groups:
        _df = groups.drop(['times', 'replicates'], axis=1)
        _df = _df.apply(np.mean, axis=0)
        data_frame[names] = _df

    # check no extraneous columns remaining
    try:
        data_frame.drop(['replicates', 'times'],
                        inplace=True,
                        axis=0)
    except KeyError:
        pass

    return data_frame


def covarFilter(infile,
                time_points,
                replicates,
                quantile):
    '''
    Filter gene list based on the distribution of the
    sums of the covariance of each gene.  This is highly
    recommended to reduce the total number of genes used
    in the dynamic time warping clustering to reduce the
    computational time.  The threshold is placed at the
    intersection of the expected and observed value
    for the given quantile.
    '''

    time_points.sort()
    time_rep_comb = [x for x in itertools.product(time_points, replicates)]
    time_cond = ro.StrVector([x[0] for x in time_rep_comb])
    rep_cond = ro.StrVector([x[1] for x in time_rep_comb])
    df = pd.read_table(infile, sep="\t", header=0, index_col=0)

    df.drop(['replicates'], inplace=True, axis=1)
    df.drop(['times'], inplace=True, axis=1)
    df = df.fillna(0.0)

    # convert data frame and import into R namespace
    # py2ri requires activation
    pandas2ri.activate()
    R.assign('diff_data', pandas2ri.py2ri(df))

    E.info("loading data frame")

    # need to be careful about column headers and transposing data frames

    R('''trans_data <- data.frame(diff_data)''')
    R('''times <- c(%s)''' % time_cond.r_repr())
    R('''replicates <- c(%s)''' % rep_cond.r_repr())

    # calculate the covariance matrix for all genes
    # sum each gene's covariance vector

    E.info("calculating sum of covariance of expression")

    R('''covar.mat <- abs(cov(trans_data))''')
    R('''sum.covar <- rowSums(covar.mat)''')
    R('''exp.covar <- abs(qnorm(ppoints(sum.covar),'''
      '''mean=mean(sum.covar), sd=sd(sum.covar)))''')
    R('''sum.covar.quant <- quantile(sum.covar)''')
    R('''exp.covar.quant <- quantile(exp.covar)''')

    E.info("filter on quantile")

    R('''filtered_genes <- names(sum.covar[sum.covar > '''
      '''sum.covar.quant[%(quantile)i]'''
      ''' & sum.covar > exp.covar.quant[%(quantile)i]])''' % locals())
    R('''filtered_frame <- data.frame(diff_data[, filtered_genes],'''
      '''times, replicates)''')

    # load data and convert to pandas object
    filtered_frame = pandas2ri.ri2py(R["filtered_frame"]).T

    return filtered_frame


def clusterPCA(infile,
               cluster_file,
               image_dir):
    '''
    PCA for each module within an experimental condition across
    the time series.
    Take PC1 as the module eigengene and return the loadings and proportion
    of variance explained for the eigengene.
    The eigengene expression across the time series is taken to be the
    value for PC1 at each timepoint as a vector.
    This is basically what WGCNA moduleEigengenes does but it
    does not recover the PC loadings.

    Warning: this script will error if there is only one
    cluster. Make sure you have more than one cluster before
    trying to perform uninformative analyses.

    Parameters:

    rpath : string
        Path of R support libraries

    '''

    header = cluster_file.split("/")[-1].split("-")[0]

    # reshape data
    R('''sink(file='sink_file.txt')''')
    R('''suppressMessages(library("reshape2"))''')
    R('''suppressMessages(library("WGCNA"))''')
    # AH: these were hard-coded paths, parameterized them to point to the
    # directory of this module's location
    R('''source("%s")''' % os.path.join(get_r_path(), "summarySE.R"))
    R('''source("%s")''' % os.path.join(get_r_path(), "clusterEigengenes.R"))
    R('''cluster_match <- read.table('%(cluster_file)s', h=T, '''
      '''row.names=1)''' % locals())
    R('''express_data <- read.table('%(infile)s', '''
      '''h=T, row.names=1, stringsAsFactors=F)''' % locals())
    R('''sink(file=NULL)''')
    R('''colnames(cluster_match) <- c("genes", "cluster")''')
    R('''express_data <- data.frame(t(express_data))''')
    R('''express_data$times <- as.numeric(as.character(express_data$times))''')
    R('''data_melt <- melt(express_data, '''
      '''id.vars=c("times", "replicates"))''')

    # sometimes data is read in as a factor/string.
    # Explicitly convert to numeric

    R('''data_melt$value <- as.numeric(as.character(data_melt$value))''')
    R('''data_sum <- summarySE(data_melt, measurevar="value", '''
      '''groupvars=c("times", "variable"))''')
    R('''data_mod <- data.frame(data_sum$times,'''
      ''' data_sum$variable, data_sum$value)''')
    R('''colnames(data_mod) <- c("times", "gene", "value")''')
    R('''data_wide <- dcast(data_mod, gene ~ times, value.var="value")''')
    R('''rownames(data_wide) <- data_wide$gene''')
    R('''times <- as.numeric(as.character(unique(express_data$times)))''')
    R('''data_wide <- data.frame(data_wide[,-1])''')
    R('''colnames(data_wide) <- times''')

    # derive module eigengenes - return a dataframe of eigengene expression
    R('''eigen_clustered <- clusterPCA(cluster_frame=cluster_match, '''
      '''expression_frame=data_wide, n=times)''')
    R('''eigen_frame <- eigenExpress(eigen_clustered, n=times)''')

    # generate loadings plot for each eigengene
    R('''eigenLoad(clusterPCA(cluster_frame=cluster_match, '''
      '''expression_frame=data_wide, n=times), image.dir="%(image_dir)s", '''
      '''condition="%(header)s")''' % locals())

    # generate expression profile plots for all eigengenes
    R('''eigenPlot(eigen_frame, image.dir="%(image_dir)s", '''
      '''condition="%(header)s")''' % locals())

    eigen_frame = pandas2ri.ri2py(R["eigen_frame"])
    eigen_frame.index = eigen_frame['cluster']
    eigen_frame.drop(['cluster'], inplace=True, axis=1)

    return eigen_frame




def conditionDESeq2(data_frame, header, alpha, res_dir):
    '''
    Perform DESeq2-based analysis of condition:time interaction
    dependent differential expression
    '''

    E.info("Differential expression testing for %s" % header)
    cols = data_frame.columns

    # py2ri requires activation
    pandas2ri.activate()
    counts = pandas2ri.py2ri(data_frame)

    des_times = ro.IntVector([x.split(".")[1] for x in cols])
    des_reps = ro.StrVector([x.split(".")[2] for x in cols])
    des_cond = ro.StrVector([x.split(".")[0] for x in cols])
    genes = ro.StrVector([x for x in data_frame.index])

    # setup counts table and design frame

    R('''suppressPackageStartupMessages(library("DESeq2"))''')
    R('''sink(file="/dev/null")''')
    R('''times <- as.factor(%s)''' % des_times.r_repr())
    R('''reps <- c(%s)''' % des_reps.r_repr())
    R('''condition <- c(%s)''' % des_cond.r_repr())
    R('''design <- data.frame(times, reps, condition)''')
    R('''counts <- data.frame(%s)''' % counts.r_repr())
    R('''genes <- c(%s)''' % genes.r_repr())
    R('''rownames(counts) <- genes''')
    R('''rownames(design) <- colnames(counts)''')

    # use DESeq() with LRT and reduced formula.  Use effect
    # size moderation

    R('''dds <- DESeqDataSetFromMatrix(countData=counts, '''
      '''colData=design, '''
      '''design=~reps + times + condition + times:condition)''')
    R('''dds <- DESeq(dds, test="LRT", '''
      '''reduced=~reps + times + condition, betaPrior=T)''')
    R('''res <- results(dds)[order(results(dds)$padj, na.last=T), ]''')
    R('''res.df <- data.frame(res)''')

    # generate dispersion and MA plots
    R('''png("%s/%s-dispersions.png")''' % (res_dir,
                                            header))
    R('''plotDispEsts(dds)''')
    R('''dev.off()''')

    R('''png("%s/%s-MAplot.png")''' % (res_dir,
                                       header))
    R('''plotMA(res, alpha=%0.3f, ylim=c(-5,5))''' % alpha)
    R('''dev.off()''')
    R('''sink(file=NULL)''')

    df = pandas2ri.ri2py(R['res.df'])

    return df


def timepointDESeq2(data_frame, header, alpha, res_dir):
    '''
    Perform DESeq2-based differential expression analysis of condition:time
    interaction
    '''

    E.info("Differential expression testing for %s" % header)
    cols = data_frame.columns

    # py2ri requires activation
    pandas2ri.activate()
    counts = pandas2ri.py2ri(data_frame)

    des_times = ro.IntVector([x.split(".")[1] for x in cols])
    des_reps = ro.StrVector([x.split(".")[2] for x in cols])
    genes = ro.StrVector([x for x in data_frame.index])

    # setup counts table and design frame

    R('''suppressPackageStartupMessages(library("DESeq2"))''')
    R('''sink(file="/dev/null")''')
    R('''times <- as.factor(%s)''' % des_times.r_repr())
    R('''reps <- c(%s)''' % des_reps.r_repr())
    R('''design <- data.frame(times, reps)''')
    R('''counts <- data.frame(%s)''' % counts.r_repr())
    R('''genes <- c(%s)''' % genes.r_repr())
    R('''rownames(counts) <- genes''')
    R('''rownames(design) <- colnames(counts)''')

    # use DESeq() with LRT and reduced formula.  Use effect
    # size moderation

    R('''dds <- DESeqDataSetFromMatrix(countData=counts, '''
      '''colData=design, '''
      '''design=~reps + times )''')
    R('''dds <- DESeq(dds, betaPrior=T)''')
    R('''res <- results(dds)[order(results(dds)$padj, na.last=T), ]''')
    R('''res.df <- data.frame(res)''')

    # generate dispersion and MA plots
    R('''png("%s/%s-dispersions.png")''' % (res_dir,
                                            header))
    R('''plotDispEsts(dds)''')
    R('''dev.off()''')

    R('''png("%s/%s-MAplot.png")''' % (res_dir,
                                       header))
    R('''plotMA(res, alpha=%0.3f, ylim=c(-5,5))''' % alpha)
    R('''dev.off()''')
    R('''sink(file=NULL)''')

    df = pandas2ri.ri2py(R['res.df'])

    return df


def genSigGenes(file_list, alpha, out_dir):
    '''
    With a list of input files (results from DESeq2) generate a
    dictionary of genes that are statistically significantly differentially
    expressed between conditions and/or time points.
    '''

    alpha = float(alpha)
    deg_dict = {}
    for infle in file_list:
        if infle.split("/")[0].split(".")[0] == "diff_condition":
            header = infle.split("/")[1].split(".")[1]
            header = header.rstrip("-diff-cond.tsv")
            header = "%s" % header

        elif infle.split("/")[0].split(".")[0] == "diff_timepoints":
            header = infle.split("/")[-1].split("-")[0]
            header = "%s_%s" % (header.split("_")[0],
                                header.split("_")[-1])

        in_df = pd.read_table(infle, sep="\t", header=0, index_col=0)
        sig_genes = in_df[in_df['padj'] <= alpha]
        deg_dict[header] = sig_genes.index.tolist()

    if file_list[0].split("/")[0].split(".")[0] == "diff_condition":
        condition = file_list[0].split("/")[1].split(".")[0]
        condition = "%s-condition" % condition

    elif file_list[0].split("/")[0].split(".")[0] == "diff_timepoints":
        condition = file_list[0].split("/")[-1].split(".")[0]
        condition = "%s-time" % condition

    drawVennDiagram(deg_dict, condition, out_dir)


def drawVennDiagram(deg_dict, header, out_dir):
    '''
    Take a dictionary of gene IDs, with keys corresponding
    to timepoints/differential expression analyses and
    generate a Venn diagram.  Maximum of 5 overlapping sets
    possible using R package:: VennDiagram.
    '''

    keys = deg_dict.keys()
    try:
        keys = sorted(keys, key=lambda x: int(x.split("_")[1].rstrip("-time")))
    except IndexError:
        pass

    venn_size = len(keys)
    R('''suppressPackageStartupMessages(library("VennDiagram"))''')
    n1 = set(deg_dict[keys[0]])
    n2 = set(deg_dict[keys[1]])
    area1 = len(n1)
    area2 = len(n2)
    n12 = len(n1.intersection(n2))

    # for Venn > 2 sets
    if venn_size == 3:
        n3 = set(deg_dict[keys[2]])
        area3 = len(n3)
        n13 = len(n1.intersection(n3))
        n23 = len(n2.intersection(n3))
        n123 = len((n1.intersection(n2)).intersection(n3))
        cat1, cat2, cat3 = keys

        R('''png("%(out_dir)s/%(header)s-venn.png", '''
          '''width=480, height=480)''' % locals())
        R('''draw.triple.venn(%(area1)d, %(area2)d, %(area3)d, '''
          '''%(n12)d, %(n23)d, %(n13)d, %(n123)d, '''
          '''c('%(cat1)s', '%(cat2)s', '%(cat3)s'), '''
          '''col=c('red', 'yellow', 'skyblue'), '''
          '''fill=c('red', 'yellow', 'skyblue'), '''
          '''margin=0.05, alpha=0.5)''' % locals())
        R('''dev.off()''')

    elif venn_size == 4:
        n3 = set(deg_dict[keys[2]])
        area3 = len(n3)
        n13 = len(n1.intersection(n3))
        n23 = len(n2.intersection(n3))
        n123 = len((n1.intersection(n2)).intersection(n3))

        n4 = set(deg_dict[keys[3]])
        area4 = len(n4)
        n14 = len(n1.intersection(n4))
        n24 = len(n2.intersection(n4))
        n34 = len(n3.intersection(n4))
        n124 = len((n1.intersection(n2)).intersection(n4))
        n134 = len((n1.intersection(n3)).intersection(n4))
        n234 = len((n2.intersection(n3)).intersection(n4))
        n1234 = len(((n1.intersection(n2)).intersection(n3)).intersection(n4))
        cat1, cat2, cat3, cat4 = keys

        R('''png("%(out_dir)s/%(header)s-venn.png",'''
          '''width=480, height=480)''' % locals())
        R('''draw.quad.venn(%(area1)d, %(area2)d, %(area3)d, %(area4)d,'''
          '''%(n12)d, %(n13)d, %(n14)d, %(n23)d, %(n24)d, %(n34)d,'''
          '''%(n123)d, %(n124)d, %(n134)d, %(n234)d, %(n1234)d,'''
          '''c('%(cat1)s', '%(cat2)s', '%(cat3)s', '%(cat4)s'), '''
          '''col=c("red", "yellow", "skyblue", "orange"), '''
          '''fill=c("red", "yellow", "skyblue", "orange"), '''
          '''margin=0.05, alpha=0.5)''' % locals())
        R('''dev.off()''')

    elif venn_size == 5:
        n3 = set(deg_dict[keys[2]])
        area3 = len(n3)
        n13 = len(n1.intersection(n3))
        n23 = len(n2.intersection(n3))
        n123 = len((n1.intersection(n2)).intersection(n3))

        n4 = set(deg_dict[keys[3]])
        area4 = len(n4)
        n14 = len(n1.intersection(n4))
        n24 = len(n2.intersection(n4))
        n34 = len(n3.intersection(n4))
        n124 = len((n1.intersection(n2)).intersection(n4))
        n134 = len((n1.intersection(n3)).intersection(n4))
        n234 = len((n2.intersection(n3)).intersection(n4))
        n1234 = len(((n1.intersection(n2)).intersection(n3)).intersection(n4))

        n5 = set(deg_dict[keys[4]])
        area5 = len(n5)
        n15 = len(n1.intersection(n5))
        n25 = len(n2.intersection(n5))
        n35 = len(n3.intersection(n5))
        n45 = len(n4.intersection(n5))
        n125 = len((n1.intersection(n2)).intersection(n5))
        n135 = len((n1.intersection(n3)).intersection(n5))
        n145 = len((n1.intersection(n4)).intersection(n5))
        n235 = len((n2.intersection(n3)).intersection(n5))
        n245 = len((n2.intersection(n4)).intersection(n5))
        n345 = len((n3.intersection(n4)).intersection(n5))
        n1235 = len(((n1.intersection(n2)).intersection(n3)).intersection(n5))
        n1245 = len(((n1.intersection(n2)).intersection(n4)).intersection(n5))
        n1345 = len(((n1.intersection(n3)).intersection(n4)).intersection(n5))
        n2345 = len(((n2.intersection(n3)).intersection(n4)).intersection(n5))
        nstep = ((n1.intersection(n2)).intersection(n3))
        n12345 = len((nstep.intersection(n4)).intersection(n5))
        cat1, cat2, cat3, cat4, cat5 = keys

        R('''png("%(out_dir)s/%(header)s-venn.png", '''
          '''height=480, width=480)''' % locals())
        R('''draw.quintuple.venn(%(area1)d, %(area2)d, %(area3)d, '''
          '''%(area4)d, %(area5)d, %(n12)d, %(n13)d, %(n14)d,'''
          '''%(n15)d, %(n23)d, %(n24)d, %(n25)d, %(n34)d, %(n35)d,'''
          '''%(n45)d, %(n123)d, %(n124)d, %(n125)d, %(n134)d,'''
          '''%(n135)d, %(n145)d, %(n234)d, %(n235)d, %(n245)d,'''
          '''%(n345)d, %(n1234)d, %(n1235)d, %(n1245)d, %(n1345)d,'''
          '''%(n2345)d, %(n12345)d, '''
          '''c('%(cat1)s', '%(cat2)s', '%(cat3)s', '%(cat4)s', '%(cat5)s'),'''
          '''col=c("red", "yellow", "skyblue", "orange", "purple"),'''
          '''fill=c("red", "yellow", "skyblue", "orange", "purple"),'''
          '''alpha=0.05, margin=0.05, cex=rep(0.8, 31))''' % locals())
        R('''dev.off()''')

    elif venn_size > 5:
        raise ValueError("Illegal venn diagram size, must be <= 5")


def maSigPro(infile,
             order_terms=1,
             fdr=0.01,
             adjust="BH",
             stepwise="backward",
             include_p=0.01,
             rsq=0.2,
             var_group="all"):
    '''
    Generate differentially expressed genes for each experimental
    condition across a time series.  Uses the bioconductor
    package maSigPro to derive a set of genes of interest.
    '''

    ref_gtf = str(infile).split("-")[1]
    data_frame = pd.read_table(infile, sep="\t", index_col=0, header=0)
    design_dict = {}

    for x in data_frame.index.values:
        sample_dict = {}
        condition = str(x).split(".")[0]
        sample_dict[condition] = 1
        sample_dict['times'] = int(str(x).split(".")[1])
        sample_dict['replicates'] = str(x).split(".")[2]
        design_dict[x] = sample_dict

    design_frame = pd.DataFrame(design_dict)
    design_frame = design_frame.T
    cols = ['times', 'replicates', condition]
    design_frame = design_frame[cols]
    design_file = "deseq.dir/%s-%s-design.tsv" % (condition, ref_gtf)
    design_frame.to_csv(design_file, sep="\t")
    data_file = "deseq.dir/%s-%s-data.tsv" % (condition, ref_gtf)
    results_file = "deseq.dir/%s-%s-maSigPro.tsv" % (condition, ref_gtf)

    # data frame columns must be in the order time-replicate-condition
    # for maSigPro
    # define the numnber of higher-order terms included in the models

    masigpro_out = "deseq.dir/maSigPro.out"

    R('''suppressMessages(library("maSigPro"))''')
    R('''input_data <- read.table('%(infile)s', sep="\t", '''
      '''h=T, row.names=1)''' % locals())
    R('''input_data <- t(input_data[0:(length(input_data)-2)])''')

    E.info("constructing experimental design matrix")

    R('''input_design <- data.matrix(read.table('%(design_file)s', '''
      '''sep="\t", h=T, row.names=1))''' % locals())
    R('''%(condition)s_mat <- make.design.matrix(input_design, '''
      '''degree = %(order_terms)i )''' % locals())
    R('''sink(file = '%(masigpro_out)s')''' % locals())

    E.info("fitting linear model for each gene with "
           "%i polynomial terms" % order_terms)

    R('''%(condition)s_fit <- p.vector(input_data, %(condition)s_mat, '''
      '''Q = %(fdr)f, MT.adjust = '%(adjust)s')''' % locals())

    # fit a linear model to each of the genes called as
    # differentially expressed
    # report genes with model R-squared > threshold
    # maSigPro gives an un-suppressable output to stdout
    # therefore sink is used to shunt this to a temporary file 'maSigPro.out'

    R('''%(condition)s_step <- T.fit(%(condition)s_fit, '''
      '''step.method='%(stepwise)s', alfa=%(include_p)f)''' % locals())

    E.info("selecting significantly differentially "
           "expressed genes at FDR=%0.3f" % fdr)

    R('''sink(file=NULL)''')
    R('''%(condition)s_sigs <- get.siggenes(%(condition)s_step, '''
      '''rsq=%(rsq)f, vars='%(var_group)s')''' % locals())
    R('''write.table(%(condition)s_sigs$sig.genes$%(condition)s$group.coeffs'''
      ''',file="deseq.dir/%(condition)s-%(ref_gtf)s-coefficients.tsv", '''
      '''sep="\t")''' % locals())
    R('''write.table(%(condition)s_sigs$sig.genes$%(condition)s$sig.pvalues,'''
      '''file="deseq.dir/%(condition)s-%(ref_gtf)s-pvalues.tsv",'''
      ''' sep="\t")''' % locals())
    R('''write.table(%(condition)s_sigs$summary, '''
      '''file='deseq.dir/%(condition)s-%(ref_gtf)s-geneids.tsv', '''
      '''sep="\t")''' % locals())
    # merge the p-value and coefficient results into a single file
    p_file = "deseq.dir/%(condition)s-%(ref_gtf)s-pvalues.tsv" % locals()
    coef_file = "deseq.dir/%s-%s-coefficients.tsv" % (condition,
                                                      ref_gtf)
    p_frame = pd.read_table(p_file, sep="\t")
    coef_frame = pd.read_table(coef_file, sep="\t")
    results_frame = pd.merge(coef_frame, p_frame,
                             how='right',
                             left_index=True,
                             right_index=True)

    results_frame.to_csv(results_file, sep="\t")

    R('''diff_genes <- data.frame(%(condition)s_fit$SELEC)''' % locals())
    diff_genes = pandas2ri.ri2py[R'diff_genes']

    return diff_genes




def splitReplicates(infile,
                    axis,
                    group_var,
                    outdir):
    '''
    Group data by replicates, new flat file for
    each.
    '''

    if axis == "row":
        axis = 1
    elif axis == "column":
        axis = 0

    inf_prefix = infile.split("/")[1].split("-")
    inf_prefix = inf_prefix[0] + "-" + inf_prefix[1]

    df = pd.read_table(infile,
                       sep="\t",
                       header=0,
                       index_col=0).T

    rep_groups = df.groupby(by=group_var,
                            axis=axis)
    for name, groups in rep_groups:
        outfile = outdir + "/" + inf_prefix + "-%s-expression.tsv" % name
        _df = groups.T
        _df.columns = _df.loc['times']
        _df.drop(['times'], axis=axis, inplace=True)
        _df.drop(['replicates'], axis=axis, inplace=True)
        _df.to_csv(outfile, sep="\t", index_label="gene_id")


def genResampleData(data_frame,
                    multiple_index,
                    replicates,
                    sample_reps,
                    times,
                    condition,
                    ref_gtf,
                    out_dir,
                    seed):
    '''
    Resample the data n-times with replacement - store
    in an sql database for ease of access
    '''

    vst_long = data_frame.T
    vst_long.index = multiple_index

    reps_dict = {}
    random.seed(seed)
    for it in range(1, replicates + 1):
        df = pd.DataFrame()
        df_list = []
        for i in times:
            k = str(random.randint(1,
                                   len(sample_reps)))
            series = vst_long.loc[str(i), 'R%s' % k]
            df_list.append((str(i), series))
        df = pd.DataFrame.from_items(df_list)
        cols = df.columns.tolist()
        cols = [int(x) for x in cols]
        cols.sort()
        cols = [str(x) for x in cols]
        df = df[cols]
        reps_dict[str(it)] = df
        table = "%s-%s-resample_%i-expression" % (condition,
                                                  ref_gtf,
                                                  it)
        seg_file = "%s/%s.tsv" % (out_dir, table)
        df.to_csv(seg_file, sep="\t")

    E.info("%i replicate datasets generated" % replicates)


def temporalCorrelate(series1, series2):
    '''
    Calculate the temporal correlation according to Chouakira & Nagabhushan
    Assumes both time series are of the same length
    '''

    series1 = list(series1)
    series2 = list(series2)

    sum_prod = []
    sum_usq = []
    sum_vsq = []
    for i in range(len(series1)-1):
        u = float(series1[i+1]) - float(series1[i])
        v = float(series2[i+1]) - float(series2[i])
        prod = u * v
        sum_prod.append(prod)
        sq_u = u**2
        sq_v = v**2
        sum_usq.append(sq_u)
        sum_vsq.append(sq_v)

    nume = sum(sum_prod)
    denom = math.sqrt(sum(sum_usq)) * math.sqrt(sum(sum_vsq))

    if denom != 0:
        return(nume/float(denom))
    else:
        return 0


def crossCorrelate(t, s, lag=0):
    '''
    Calculate the cross-correlation of two timeseries, s and t.
    Return the normalized correlation value at lag=n.
    Uses numpy.correlate; default is to return lag=0.
    TODO: return multiple lags?
    '''

    t_mean = np.mean(t)
    s_mean = np.mean(s)
    t_std = np.std(t)
    s_std = np.std(s)
    len_t = len(t)

    t_norm = [((x - t_mean)/(t_std * len_t)) for x in t]
    s_norm = [((y - s_mean)/s_std) for y in s]

    if lag == 0:
        xcorr = np.correlate(t_norm, s_norm)
    elif lag != 0:
        xcorr = np.correlate(t_norm, s_norm, mode=2)[len_t - 1 + lag]

    return xcorr


def adaptiveTune(value, k):
    '''
    Calculate the adaptive tuning function from Chouakira & Nagabhushan
    '''

    if k == 0:
        return 1.0
    else:
        return (2/(1 + math.exp(k*abs(value))))


def dtwWrapper(data, rows, columns, k):
    '''
    wrapper function for dynamic time warping.
    includes use of exponential adaptive tuning function
    with temporal correlation if k > 0
    '''

    # not explicitly called, but needs to be in R environment
    DTW = importr("dtw")

    # create a data frame of zeros of size number of ids x number of ids
    # fill it with the calculated distance metric for each pair wise comparison

    df_ = pd.DataFrame(index=rows,
                       columns=columns)
    df_ = df_.fillna(0.0).astype(np.float64)

    # fill the array with dtw-distance values
    pandas2ri.activate()

    for i in rows:
        E.info("DTW %s" % i)
        for j in columns:
            series1 = data.loc[i].values.tolist()
            series2 = data.loc[j].values.tolist()
            DTW_value = (R.dtw(series1,
                               series2)).rx('distance')[0][0]
            cort_value = temporalCorrelate(series1, series2)
            tuned_value = adaptiveTune(cort_value, k)
            time_dist = DTW_value * tuned_value
            df_.loc[i][j] = float(time_dist)
            df_[j][i] = float(time_dist)

    return df_


def correlateDistanceMetric(data, rows, columns, method, lag=0):
    '''
    wrapper for correlation coefficients as distance metrics
    for time-series clustering.
    Use either temporal correlation (analagous to template matching)
    or normalised cross correlation.
    '''

    # create blank (all 0's) dataframe to fill with correlation values

    df_ = pd.DataFrame(index=rows,
                       columns=columns)
    df_ = df_.fillna(0.0)

    if method == "cross-correlate":
        for i in rows:
            E.info("cross-correlation %s" % i)
            for j in columns:
                series1 = data.loc[i].values.tolist()
                series2 = data.loc[j].values.tolist()
                corr = crossCorrelate(series1, series2, lag=lag)
                df_.loc[i][j] = 1.0 - abs(corr)
                df_[j][i] = 1.0 - abs(corr)

    elif method == "temporal-correlate":
        for i in rows:
            E.info("temporal correlation %s" % i)
            for j in columns:
                series1 = data.loc[i].tolist()
                series2 = data.loc[j].tolist()
                corr = temporalCorrelate(series1, series2)
                df_.loc[i][j] = 1.0 - abs(corr)
                df_[j][i] = 1.0 - abs(corr)

    return df_


def splitFiles(infile, nchunks, out_dir):
    '''
    Give files names based on splitting into an arbitrary number of chunks
    '''

    df = pd.read_table(infile, sep="\t", header=0, index_col=0)
    total = len(df.index.tolist())

    # split into aribitrary number of chunks, or arbitrary chunk size?
    # small n bad for large input size, large n bad for small input size
    # set min/max chunk size, e.g. 100 genes minimum, 500 maximum?

    if total/nchunks < 100:
        step = 100
        E.warn("too few genes in each chunk, resetting to 100 genes per chunk")
    elif total/nchunks > 500:
        step = 500
        E.warn("too many genes per chunk, resetting to 500 genes per chunk")
    else:
        step = total/nchunks
        E.info("chunking input file into %i chunks" % step)

    file_pattern = infile.split("/")[1].rstrip("-expression.tsv")
    idx = 0
    for i in range(step, total, step):
        start = "%s" % idx
        end = "%s" % i
        file_name = "%s/%s-%s_%s-split.tsv" % (out_dir,
                                               file_pattern,
                                               start,
                                               end)
        with open(file_name, "w") as file_handle:
            file_handle.write(file_name + "\n")
        idx = i

    # final file
    start = "%s" % idx
    end = "%s" % total
    file_name = "%s/%s-%s_%s-split.tsv" % (out_dir,
                                           file_pattern,
                                           start,
                                           end)
    with open(file_name, "w") as file_handle:
        file_handle.write(file_name + "\n")


def mergeFiles(file_list, outfile):
    '''
    Merge files after split-transform operations
    '''

    # sort list by starting index of file
    res_list = sorted(file_list,
                      key=lambda x: int(x.split("/")[-1].split("-")[3].split("_")[0]))

    # merge files using pandas data frame merge method
    full_frame = pd.read_table(res_list[0], sep="\t", index_col=0, header=0)
    res_list.remove(res_list[0])
    for fle in res_list:
        df = pd.read_table(fle, sep="\t", index_col=0, header=0)
        full_frame = pd.merge(left=full_frame,
                              right=df,
                              how='inner',
                              left_index=True,
                              right_index=True)

    full_frame.to_csv(outfile, sep="\t")


def treeCutting(infile,
                expression_file,
                cluster_file,
                cluster_algorithm,
                deepsplit=False):
    '''
    Use dynamic tree cutting to derive clusters for each
    resampled distance matrix
    '''
    wgcna_out = "/dev/null"

    E.info("loading distance matrix")

    df = pd.read_table(infile, sep="\t",
                       header=0, index_col=0)
    df = df.fillna(0.0)
    genes = df.index
    genes_r = ro.StrVector([g for g in genes])

    # py2ri requires activation
    pandas2ri.activate()
    rdf = pandas2ri.py2ri(df)

    R.assign("distance_data", rdf)
    R.assign("gene_ids", genes_r)

    R('''sink(file='%(wgcna_out)s')''' % locals())
    R('''suppressPackageStartupMessages(library("WGCNA"))''')
    R('''suppressPackageStartupMessages(library("flashClust"))''')
    E.info("clustering data by %s linkage" % cluster_algorithm)
    R('''rownames(distance_data) <- gene_ids''')
    R('''clustering <- flashClust(as.dist(distance_data),'''
      ''' method='%(cluster_algorithm)s')''' % locals())
    if deepsplit:
        R('''cluster_cut <- cutreeDynamic(dendro=clustering, '''
          '''minClusterSize=50, deepSplit=T)''')
    else:
        R('''cluster_cut <- cutreeDynamic(dendro=clustering, '''
          '''minClusterSize=50, deepSplit=F)''')

    R('''color_cut <- labels2colors(cluster_cut)''')
    R('''write.table(color_cut, file = '%(cluster_file)s','''
      '''sep="\t")''' % locals())
    R('''cluster_matched <- data.frame(cbind(rownames(distance_data),'''
      '''color_cut))''')
    R('''colnames(cluster_matched) = c("gene_id", "cluster")''')
    R('''cluster_matched <- data.frame(cluster_matched$gene_id,'''
      '''cluster_matched$cluster)''')
    R('''sink(file=NULL)''')

    cluster_frame = pandas2ri.ri2py(R["cluster_matched"])
    cluster_frame.columns = ['gene_id', 'cluster']
    cluster_frame.index = cluster_frame['gene_id']
    cluster_frame.drop(['gene_id'], inplace=True, axis=1)

    return cluster_frame


def clusterAverage(file_list):
    '''
    Average distance measures across replicates
    '''

    # map replicate number on to dataframe as identifier
    # used to aggregate over replicates for each gene id
    # assumes filename is condition-reference-replicate

    df_dict = {}
    for fle in file_list:
        f = fle.split("/")[-1]
        rep = f.split("-")[2]
        _df = pd.read_table(fle, sep="\t",
                            header=0, index_col=0)
        df_dict[rep] = _df

    # group concatenated dataframes by gene_id
    concat_df = pd.concat(df_dict)
    group_df = concat_df.groupby(level=1)

    # aggregate/summarise distance metric over replicates for each gene_id
    agg_dict = {}
    for names, groups in group_df:
        agg_dict[names] = np.mean(groups, axis=0)

    agg_df = pd.DataFrame(agg_dict)
    return agg_df


def clusterAgreement(infile):
    '''
    calculate co-occurence of genes within resampled clusters
    '''
    # read in aggregated cluster assignment file, genes as rows,
    # iterations as columns

    df = pd.read_table(infile, sep="\t", header=0, index_col=0)
    genes = df.index.values

    # instantiate an empy matrix to count the number of times each
    # gene appears with others at each iteration

    dmat = pd.DataFrame(index=genes,
                        columns=genes)
    dmat = dmat.fillna(0)

    # generate all pair-wise combinations of genes, index into dmat using these
    reps = df.columns.values

    # alternative, faster code for consensus clustering - might be able to
    # improve it by using less nested for loops

    # generate a set for each cluster that contains the genes belonging to
    # each cluster count the number of times two genes occur in a cluster
    # and add to the dataframe repeat for each resampling iteration
    # cluster sets are generated at every iteration - time improvement over
    # directly accessing and storing as two dataframes is a factor of 7-8

    for i in reps:
        # from the input dataframe generate a list of sets, one set for each
        # cluster in that resampling iteration.
        # repeat for every resampling iteration.

        clusters = set(df[i].values.tolist())
        cluster_dict = {}
        for col in clusters:
            cluster_dict[col] = []
        for gene in genes:
            k_gene = df[i][gene]
            cluster_dict[k_gene].append(gene)
        rep_list = []

        # for each cluster add all the genes with that cluster ID to the set
        # add all of the cluster sets to a list container

        for col in clusters:
            col_set = set()
            clust_col = cluster_dict[col]
            gene_members = itertools.combinations_with_replacement(clust_col,
                                                                   2)
            col_set.add(gene_members)
            rep_list.append(col_set)

        # count if two genes occur in the same cluster, adding to both sides
        # of the symmetrical dataframe

        for cluster_set in rep_list:
            for combs in cluster_set:
                for x in combs:
                    if x[0] == x[1]:
                        dmat[x[0]][x[1]] += 1
                    else:
                        dmat[x[0]][x[1]] += 1
                        dmat[x[1]][x[0]] += 1

    # calculate the proportion of co-occurences

    prob = lambda x: x/float(len(reps))

    probs_df = dmat.applymap(prob)

    return probs_df


def consensusClustering(infile,
                        cutHeight,
                        cluster_algorithm,
                        min_size=30,
                        deepsplit=False):
    '''
    hierachichal clustering based on gene-cluster correlation across
    resampled datasets.  cut tree based with dynamic tree cut
    TODO: change this to cutHeight?  i.e. 0.2 = 80% clustering
    agreement OR use dynamic tree cut without deepsplit.
    '''
    condition = infile.split("/")[1].split("-")[0]
    wgcna_out = "tmp.dir/consensus-WGCNA.out"

    R('''sink(file='%(wgcna_out)s')''' % locals())
    R('''suppressMessages(library("WGCNA"))''')
    R('''suppressMessages(library("flashClust"))''')

    E.info("loading distance matrix")

    df = pd.read_table(infile, sep="\t", header=0, index_col=0)
    labels = df.index.tolist()
    labels_r = ro.StrVector([l for l in labels])

    # py2ri requires activation
    pandas2ri.activate()
    df_r = pandas2ri.py2ri(df)

    R.assign("distance.frame", df_r)
    R.assign("labels", labels_r)

    # large matricies/distance objects may need more
    # memory - allocate 1GB
    R('''memory.limit(10000)''')
    R('''rownames(distance.frame) <- labels''')
    R('''distance_data <- data.matrix(distance.frame)''')

    E.info("clustering data by %s linkage" % cluster_algorithm)

    R('''clustering <- flashClust(as.dist(1-distance_data),'''
      '''method='%(cluster_algorithm)s')''' % locals())

    if cutHeight > float(0.01):
        R('''cluster_cut <- cutreeStatic(dendro=clustering, '''
          '''minSize=%(min_size)i, cutHeight=%(cutHeight)s)''' % locals())

    elif deepsplit:
        R('''cluster_cut <- cutreeDynamic(dendro=clustering, '''
          '''deepSplit=T, minClusterSize=%(min_size)i)''' % locals())
    else:
        R('''cluster_cut <- cutreeDynamic(dendro=clustering, '''
          '''deepSplit=F, minClusterSize=%(min_size)i)''' % locals())

    R('''color_cut <- labels2colors(cluster_cut)''')
    R('''cluster_matched <- data.frame(cbind(rownames(distance_data),'''
      '''color_cut))''')
    R('''colnames(cluster_matched) = c("gene_id", "cluster")''')
    R('''cluster_matched <- data.frame(cluster_matched$gene_id,'''
      '''cluster_matched$cluster)''')

    # plot and save dendrogram of clustering
    # AH: disabled, requires plots.dir to exist which might not be the case
    # AH: and thus causes this method to fail. Path names need to be parameterizable.
    # R('''png("plots.dir/%(condition)s-dendrogram-consensus_clustering.png")'''
    #   % locals())
    # R('''plotDendroAndColors(dendro=clustering, colors=color_cut,'''
    #   '''groupLabels="Dynamic tree cut",'''
    #   '''dendroLabels=F, addGuide=T, guideHang=0.05, '''
    #   '''hang=0.03, main="%(condition)s")''' % locals())
    # R('''dev.off()''')
    # R('''sink(file=NULL)''')
    cluster_frame = pandas2ri.ri2py(R["cluster_matched"])

    return cluster_frame



'''
Tophat.py - working with tophat/cufflinks output files
======================================================

:Tags: Python

Code
----

'''

import re
import collections
import itertools

from CGAT import IOTools as IOTools


class CuffCompareValues:

    def __init__(self, vals):

        assert len(vals) == 4
        try:
            self.sn = float(vals[0])
        except ValueError:
            self.sn = None

        try:
            self.sp = float(vals[1])
        except ValueError:
            self.sp = None

        try:
            self.fsn = float(vals[2])
        except ValueError:
            self.fsn = None

        try:
            self.fsp = float(vals[3])
        except ValueError:
            self.fsp = None

    def __str__(self):
        return "\t".join(map(IOTools.val2str,
                             (self.sn,
                              self.sp,
                              self.fsn,
                              self.fsp)))

    @classmethod
    def getHeaders(cls):
        return ("sn", "sp", "fsn", "fsp")


class CuffCompareResult:

    def __init__(self):
        (self.baselevel,
         self.exonlevel,
         self.intronlevel,
         self.intronchainlevel,
         self.transcriptlevel,
         self.locuslevel,
         self.missedexons_counts,
         self.missedexons_total,
         self.novelexons_counts,
         self.novelexons_total,
         self.missedintrons_counts,
         self.missedintrons_total,
         self.novelintrons_counts,
         self.novelintrons_total,
         self.missedloci_counts,
         self.missedloci_total,
         self.novelloci_counts,
         self.novelloci_total,
         self.matchingintronchains_count,
         self.matchingloci_count,
         self.query,
         self.query_loci,
         self.query_multi_exon,
         self.reference,
         self.reference_loci,
         self.reference_multi_exon,
         self.loci_multi_exon,
         self.loci_transcripts,
         ) = [None] * 28

    def fromLines(self, lines):
        '''parse from cuffcompare output.'''
        self.is_empty = True

        for line in lines:
            if line.startswith("#"):
                d = line[1:-1].strip()
                if d.startswith("Query"):
                    self.query, self.query_loci, self.query_multi_exon = list(map(int, re.match(
                        "Query mRNAs :\s+(\d+)\s+in\s+(\d+)\s+loci\s+\((\d+)", d).groups()))
                elif d.startswith("Reference"):
                    self.reference, self.reference_loci, self.reference_multi_exon = list(map(int, re.match(
                        "Reference mRNAs :\s+(\d+)\s+in\s+(\d+)\s+loci\s+\((\d+)", d).groups()))
                elif d.startswith("("):
                    self.loci_multi_exon, self.loci_transcripts = re.match(
                        "\((\d+) multi-transcript loci, ~(\S+)", d).groups()
                    self.loci_multi_exon = int(self.loci_multi_exon)
                    self.loci_transcripts = float(self.loci_transcripts)
                continue

            line = line[:-1].strip()
            if not line:
                continue

            try:
                tag, data = line.split(":")
            except ValueError:
                raise ValueError("parsing error in line %s" % line)

            tag = re.sub("\s", "", tag).lower()

            if tag.startswith("novel") or tag.startswith("missed"):

                counts, total = list(map(
                    int, re.match("\s+(\d+)/(\d+)", data).groups()))
                setattr(self, "%s_counts" % tag, counts)
                setattr(self, "%s_total" % tag, total)

            # NICK FIX: This output changed from the previous version of cufflinks. not sure if it means the same
            # but nevertheless the new output will be put into sphinxreport
            elif tag.startswith("matching"):
                counts = int(data)
                setattr(self, "%s_count" % tag, counts)

            elif tag.startswith("total"):
                # stop at total union across all super-loci
                break
            else:
                values = data.strip().split()
                setattr(self, tag, CuffCompareValues(values))
                self.is_empty = False

    def __str__(self):
        return "\t".join(map(str, (
            self.baselevel,
            self.exonlevel,
            self.intronlevel,
            self.intronchainlevel,
            self.transcriptlevel,
            self.locuslevel,
            self.missedexons_counts,
            self.missedexons_total,
            self.novelexons_counts,
            self.novelexons_total,
            self.missedintrons_counts,
            self.missedintrons_total,
            self.novelintrons_counts,
            self.novelintrons_total,
            self.missedloci_counts,
            self.missedloci_total,
            self.novelloci_counts,
            self.novelloci_total,
            self.matchingintronchains_count,
            self.matchingloci_count,
            self.query,
            self.query_loci,
            self.query_multi_exon,
            self.reference,
            self.reference_loci,
            self.reference_multi_exon,
            self.loci_multi_exon,
            self.loci_transcripts
        )))

    @classmethod
    def getHeaders(cls):
        m = ("baselevel", "exonlevel", "intronlevel", "intronchainlevel",
             "transcriptlevel", "locuslevel")
        a = ("missed", "novel")
        b = ("exons", "introns", "loci")
        c = ("counts", "total")
        return [ "%s_%s" % (x, y) for x, y in itertools.product( m, CuffCompareValues.getHeaders() ) ] +\
            [ "%s%s_%s" % (x, y, z) for x, y, z in itertools.product( a, b, c) ] +\
            ["matchingintronchains_count", "matchingloci_count"] +\
            ["query", "query_loci", "query_multi_exon",
             "reference", "reference_loci", "reference_multi_exon",
             "loci", "loci_multi_exon"]


def parseTranscriptComparison(infile):
    '''read cufflinks 1.0.3 output in infile stream.

    returns a two-level dictionary mapping with levels track and contig.
    '''

    firstline = infile.readline()
    result = collections.defaultdict(dict)
    tracks = []

    def __blocker(infile):
        # 'Summary for dataset'
        # comes after 'Genomic sequence'.

        blocks = {}
        contig, block = None, []
        # whether to yield
        y = False
        for line in infile:

            if line.startswith("#> Genomic sequence"):
                if block:
                    blocks[contig] = block
                if y:
                    yield dataset, blocks
                    blocks = {}
                    y = False

                contig = re.match(
                    "#> Genomic sequence: (\S+)", line).groups()[0]
                block = []
                continue
            elif line.startswith("#= Summary for dataset:"):
                if block:
                    blocks[contig] = block
                if y:
                    yield dataset, blocks
                    blocks = {}
                    y = False

                dataset = re.match(
                    "#= Summary for dataset: (\S+)", line).groups()[0]
                contig = "all"
                block = []

                y = True
                continue

            block.append(line)

        blocks[contig] = block
        yield dataset, blocks

    for track, blocks in __blocker(infile):
        print(track)
        tracks.append(track)
        for contig, block in blocks.items():
            r = CuffCompareResult()

            r.fromLines(block)
            result[track][contig] = r

    return tracks, result

Locus = collections.namedtuple(
    "Locus", "locus_id contig strand start end transcript_ids transcripts")
Tracking = collections.namedtuple(
    "Tracking", "transfrag_id locus_id ref_gene_id ref_transcript_id code transcripts")
TranscriptInfo = collections.namedtuple(
    "Transfrag", "gene_id transcript_id fmi fpkm conf_lo conf_hi cov len")


def iterate_tracking(infile):
    '''parse .tracking output file from cuffcompare

    returns iterator with list of loci.
    '''
    for line in infile:
        data = line[:-1].split("\t")
        try:
            transfrag_id, locus_id, transcript_id, code = [
                x.strip() for x in data[:4]]
        except ValueError:
            raise ValueError(
                "parsing error: expected transfrag_id, locus_id, transcript_id, code, got %s" % str(data))

        if transcript_id == "-":
            ref_gene_id, ref_transcript_id = "", ""
        else:
            ref_gene_id, ref_transcript_id = transcript_id.split("|")

        transcripts = []
        for c in data[4:]:
            if c == "-":
                transcripts.append(None)
            else:
                # there can be multiple transcripts
                for cc in c.split(","):
                    try:
                        (gene_id, transcript_id, fmi, fpkm, conf_lo,
                         conf_hi, cov, length) = cc.split("|")
                    except ValueError:
                        raise ValueError("parsing error for field '%s'" % cc)
                    (fpkm, conf_lo, conf_hi, cov) = list(map(
                        float, (fpkm, conf_lo, conf_hi, cov)))
                    if length == "-":
                        length = 0
                    (fmi, length) = list(map(int, (fmi, length)))

                    transcripts.append(TranscriptInfo._make(
                        (gene_id, transcript_id, fmi, fpkm, conf_lo, conf_hi, cov, length)))

        yield Tracking._make((transfrag_id, locus_id, ref_gene_id, ref_transcript_id, code, transcripts))


def iterate_locus(infile):
    '''parse .loci output file from cuffcompare

    returns iterator with list of loci.
    '''

    for line in infile:
        data = line[:-1].split("\t")
        locus_id, pos = data[:2]

        contig, strand, start, end = re.match(
            "(\S+)\[(\S*)\](\d+)-(\d+)", pos).groups()
        start, end = list(map(int, (start, end)))

        # some [] contains the 0 byte, convert to empty field
        if strand not in "+-":
            strand = ""
        transcripts = []
        for c in data[2:]:
            if c == "-":
                transcripts.append(())
            else:
                transcripts.append(c.split(","))

        yield Locus._make((locus_id, contig, strand, start, end, transcripts[1], transcripts[1:]))



"""
bam_vs_bam.py - compute coverage correlation between bam files
==============================================================

:Tags: Genomics NGS BAM Comparison

Purpose
-------

Compare per base coverage between two :term:`bam` formatted files.

Usage
-----

Example::

   python bam_vs_bam.py in1.bam in2.bam

This command generates a tab delimited output with columns chromosome,
base coordinate, number of overlapping reads in in1.bam, and number of
overlapping reads in in2.bam.

Type::

   python bam_vs_bam.py --help

for command line help.

Documentation
-------------

This tools allows users to compare the per base coverage between
two BAM files. The output includes all bases in the supplied reference
fasta except those with no coverage in the input BAMs.

At present the --interval or -i option has not been implemented.

Command line options
--------------------

``--regex-identifier``
    supply a regex to extract an identifier from the filenames.
    defualts to using the filename

"""

import sys
import re
import pysam
import CGAT.Experiment as E


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-i", "--intervals-bed-file", dest="filename_intervals",
                      type="string",
                      help="filename with intervals to use "
                      "[%default].")

    parser.add_option("-e", "--regex-identifier", dest="regex_identifier",
                      type="string",
                      help="regular expression to extract identifier from "
                      "filename [%default].")

    parser.set_defaults(
        filename_intervals=None,
        regex_identifier="(.*)",
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if len(args) < 1:
        raise ValueError("please supply at least two BAM files.")

    samfiles = []
    for f in args:
        samfiles.append(pysam.AlignmentFile(f, "rb"))

    if options.filename_intervals:
        raise NotImplementedError(
            "It is not yet possible to specify intervals of interest.\
            Repeat command without intervals option.")

    titles = [re.search(options.regex_identifier, x).groups()[0] for x in args]

    options.stdout.write("contig\tpos\t%s\n" % "\t".join(titles))

    ninput, nskipped, noutput = 0, 0, 0
    contigs = samfiles[0].references

    for contig in contigs:

        missing_contig = False

        positions = {}

        # lazy way: use dictionary
        for x, f in enumerate(samfiles):
            try:
                i = f.pileup(contig)
            except ValueError:
                missing_contig = True
                break

            for v in i:
                vp = v.pos
                if vp in positions:
                    positions[vp].append(v.n)
                else:
                    positions[vp] = [0] * x + [v.n]

            # fill with 0 those not touched in this file
            for p in list(positions.keys()):
                if len(positions[p]) <= x:
                    positions[p].append(0)

        if missing_contig:
            nskipped += 1
            continue

        noutput += 1
        for pos in sorted(positions.keys()):
            vals = positions[pos]
            options.stdout.write("%s\t%i\t%s\n" % (contig, pos,
                                                   "\t".join(map(str, vals))))

    E.info("ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped))

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
bed2fasta.py - get sequences from bed file
==========================================

:Tags: Genomics Intervals Sequences Conversion BED FASTA



Purpose
-------

This script outputs nucleotide sequences for intervals within
a :term:`bed` formatted file using a corresponding genome file.

Usage
-----

A required input to bed2fasta.py is a CGAT indexed genome. To obtain an
idexed human reference genome we would type

Example::
   cat hg19.fasta | index_fasta.py hg19 > hg19.log

This file would then serve as the --genome-file when we wish to extract
sequences from a :term:`bed` formatted file.


For example we could now type::

   cat in.bed | python bed2fasta.py --genome-file hg19 > out.fasta

Where we take a set of genomic intervals (e.g. from a human ChIP-seq experiment)
and output their respective nucleotide sequences.


Type::

   python bed2fasta.py --help

for command line help.

Command line options
--------------------

'''
import sys
import CGAT.Experiment as E
import CGAT.Bed as Bed
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Masker as Masker


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genomic sequence to retrieve "
                      "sequences from.")

    parser.add_option("-m", "--masker", dest="masker", type="choice",
                      choices=("dust", "dustmasker", "softmask", "none"),
                      help="apply masker to mask output sequences "
                      "[%default].")

    parser.add_option("--output-mode", dest="output_mode", type="choice",
                      choices=("intervals", "leftright", "segments"),
                      help="what to output. "
                      "'intervals' generates a single sequence for "
                      "each bed interval. 'leftright' generates two "
                      "sequences, one in each direction, for each bed "
                      "interval. 'segments' can be used to output "
                      "sequence from bed12 files so that sequence only covers "
                      "the segements [%default]")

    parser.add_option("--min-sequence-length", dest="min_length", type="int",
                      help="require a minimum sequence length [%default]")

    parser.add_option("--max-sequence-length", dest="max_length", type="int",
                      help="require a maximum sequence length [%default]")

    parser.add_option(
        "--extend-at", dest="extend_at", type="choice",
        choices=("none", "3", "5", "both", "3only", "5only"),
        help="extend at 3', 5' or both or no ends. If 3only or 5only "
        "are set, only the added sequence is returned [default=%default]")

    parser.add_option(
        "--extend-by", dest="extend_by", type="int",
        help="extend by # bases [default=%default]")

    parser.add_option(
        "--use-strand", dest="ignore_strand",
        action="store_false",
        help="use strand information and return reverse complement "
        "on intervals located on the negative strand. "
        "[default=%default]")

    parser.set_defaults(
        genome_file=None,
        masker=None,
        output_mode="intervals",
        min_length=0,
        max_length=0,
        extend_at=None,
        extend_by=100,
        ignore_strand=True,
    )

    (options, args) = E.Start(parser)

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
        contigs = fasta.getContigSizes()
        fasta.setConverter(IndexedFasta.getConverter("zero-both-open"))

    counter = E.Counter()
    ids, seqs = [], []

    E.info("collecting sequences")
    for bed in Bed.setName(Bed.iterator(options.stdin)):
        counter.input += 1

        lcontig = fasta.getLength(bed.contig)

        if options.ignore_strand:
            strand = "+"
        else:
            strand = bed.strand

        if options.output_mode == "segments" and bed.columns == 12:
            ids.append("%s %s:%i..%i (%s) %s %s" %
                       (bed.name, bed.contig, bed.start, bed.end, strand,
                        bed["blockSizes"], bed["blockStarts"]))
            seg_seqs = [fasta.getSequence(bed.contig, strand, start, end)
                        for start, end in bed.toIntervals()]
            seqs.append("".join(seg_seqs))

        elif (options.output_mode == "intervals" or
              options.output_mode == "segments"):
            ids.append("%s %s:%i..%i (%s)" %
                       (bed.name, bed.contig, bed.start, bed.end, strand))
            seqs.append(
                fasta.getSequence(bed.contig, strand, bed.start, bed.end))

        elif options.output_mode == "leftright":
            l = bed.end - bed.start

            start, end = max(0, bed.start - l), bed.end - l
            ids.append("%s_l %s:%i..%i (%s)" %
                       (bed.name, bed.contig, start, end, strand))
            seqs.append(fasta.getSequence(bed.contig, strand, start, end))

            start, end = bed.start + l, min(lcontig, bed.end + l)
            ids.append("%s_r %s:%i..%i (%s)" %
                       (bed.name, bed.contig, start, end, strand))
            seqs.append(fasta.getSequence(bed.contig, strand, start, end))

    E.info("collected %i sequences" % len(seqs))

    masked = Masker.maskSequences(seqs, options.masker)
    options.stdout.write(
        "\n".join([">%s\n%s" % (x, y) for x, y in zip(ids, masked)]) + "\n")

    E.info("masked %i sequences" % len(seqs))

    counter.output = len(seqs)

    E.info("%s" % counter)

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Bioprospector.py - Parser for Bioprospector output
==================================================

:Tags: Python
"""

import sys
import re
import collections

import weblogolib
import corebio.seq

Match = collections.namedtuple("Match",
                               '''id, site, strand, start, end,
                                  width1, width2, sequence''')

Motif = collections.namedtuple("Motif",
                               '''id, pattern, width1, width2,
                                min_gap, max_gap,
                                score, sites, matches''')


def parse(infile):
    '''parse Bioprospector output.'''

    motifs = []

    def iterate_blocks(infile):
        lines = []
        keep = False
        for line in infile:
            if line.startswith("Motif #"):
                if lines:
                    yield lines
                keep = True
                lines = []

            if line.startswith("*****"):
                continue

            if keep:
                lines.append(line)
        if lines:
            yield(lines)

    for lines in iterate_blocks(infile):
        motif_id, pattern = re.match(
            "Motif \#(\d+): \((.+)\)", lines[0]).groups()
        widths, gaps, score, sites = re.match(
            "Width \((.*)\); Gap \[(.*)\]; MotifScore (.*); Sites (.*)", lines[1]).groups()
        width1, width2 = list(map(int, widths.split(",")))
        min_gap, max_gap = list(map(int, gaps.split(",")))

        sites = int(sites)
        start = end = 3
        while not lines[end].startswith(">"):
            end += 1

        block = lines[start:end]

        def convert(coords, width):
            if coords[0] == "f":
                strand, start = "+", int(coords[1]) - 1
            else:
                strand, start = "-", int(coords[1]) - width
            return strand, start

        sequences = []
        for x in range(end, len(lines), 2):
            if lines[x].strip() == "":
                break
            assert lines[x].startswith(">"), lines[x]
            id, lseq, site_id, coords = re.match(
                ">(\S+).*len\s(\d+).*site \#(\d+)\s*(.*)", lines[x][:-1]).groups()
            lseq, site_id = list(map(int, (lseq, site_id)))
            coords = re.split("\s", coords)
            if len(coords) == 2:
                strand1, start1 = convert(coords[:2], width1)
                strand = strand1
                start, end = start, start + width1
            elif len(coords) == 4:
                strand1, start1 = convert(coords[:2], width1)
                strand2, start2 = convert(coords[2:], width2)
                start, end = min(start1, start2), max(
                    start1 + width1, start2 + width2)
                strand = strand1 + strand2

            seq = lines[x + 1][:-1]
            sequences.append(Match._make((id, site_id,
                                          strand,
                                          start, end,
                                          width1, width2,
                                          seq)))

        motif = Motif._make((motif_id, pattern,
                             width1, width2,
                             min_gap, max_gap,
                             score,
                             sites, sequences))
        motifs.append(motif)

    return motifs


def build_logo(sequences, outfilename):

    seqs = corebio.seq.SeqList(alphabet=corebio.seq.dna_alphabet)
    for sequence in sequences:
        seqs.append(corebio.seq.dna(re.sub("\s", "-", sequence)))

    data = weblogolib.LogoData.from_seqs(seqs)
    options = weblogolib.LogoOptions()
    options.color_scheme = weblogolib.classic

    options.title = 'A Logo Title'
    format = weblogolib.LogoFormat(data, options)
    fout = open(outfilename, 'w')
    weblogolib.png_formatter(data, format, fout)

if __name__ == "__main__":

    results = parse(sys.stdin)

    outname = "test%i.eps"

    from . import MAST
    for x, motifs in enumerate(results):
        build_logo([y.sequence for y in motifs.matches], outname % x)
        print(motifs.score, motifs.sites)

        outfile = open("test%i.motif" % x, "w")
        MAST.sequences2motif(outfile,
                             [y.sequence for y in motifs.matches])
        outfile.close()

        outfile = open("test%i.tomtom" % x, "w")
        MAST.sequences2motif(outfile,
                             [y.sequence for y in motifs.matches],
                             format="TOMTOM")
        outfile.close()



"""
diff_bed.py - count differences between several bed files
=========================================================

:Tags: Genomics Intervals BED Comparison

Purpose
-------

Compute overlap statistics between multiple bed files. For each pairwise
comparison, this script outputs the number of intervals (exons) and
bases overlapping.

Using the ``--update`` option, a table can be incrementally updated with
additional comparisons.

The strand of intervals is ignored in comparisons.

+--------------+----------------------------------+
|*Column*      |*Content*                         |
+--------------+----------------------------------+
|set           |Name of the set                   |
+--------------+----------------------------------+
|nexons_total  |number of intervals in set        |
+--------------+----------------------------------+
|nexons_ovl    |number of intervals overlapping   |
+--------------+----------------------------------+
|nexons_unique |number of unique intervals        |
+--------------+----------------------------------+
|nbases_total  |number of bases in gene set       |
+--------------+----------------------------------+
|nbases_ovl    |number of bases overlapping       |
+--------------+----------------------------------+
|nbases_unique |number of unique bases            |
+--------------+----------------------------------+

Usage
-----

For example::

   python diff_bed.py *.bed.gz > out.tsv

To update results from a previous run, type::

   python diff_bed.py --update=out.tsv *.bed.gz > new.tsv

Type::

   python diff_bed.py --help

for command line help.

Command line options
--------------------

"""

import sys
import re
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Bed as Bed
import numpy


class Counter:

    mPercentFormat = "%5.2f"

    def __init__(self):
        pass

    def getHeader(self):
        h = []
        for a in ("exons", "bases"):
            for b in ("total", "ovl", "unique"):
                for c in ("1", "2"):
                    h.append("n" + a + "_" + b + c)
        for a in ("exons", "bases"):
            for b in ("ovl", "unique"):
                for c in ("1", "2"):
                    h.append("p" + a + "_" + b + c)

        return "\t".join(h)

    @E.cachedmethod
    def buildIndex(self, filename):
        return Bed.readAndIndex(IOTools.openFile(filename, "r"))

    def _count(self, filename, idx):
        '''count filename against idx.'''

        overlapping_genes = set()
        genes = set()

        # iterate over exons
        infile = IOTools.openFile(filename, "r")
        it = Bed.bed_iterator(infile)

        nexons, nexons_overlapping = 0, 0
        nbases, nbases_overlapping = 0, 0
        for this in it:
            nexons += 1
            nbases += this.end - this.start

            try:
                intervals = list(
                    idx[this.contig].find(max(0, this.start), this.end))
            except KeyError:
                continue
            except Exception as msg:
                raise Exception(
                    "error while processing %s, msg=%s" % (filename, msg))
            if len(intervals) == 0:
                continue

            nexons_overlapping += 1
            start, end = this.start, this.end
            counts = numpy.zeros(end - start, numpy.int)
            for other_start, other_end, other_value in intervals:
                for x in range(max(start, other_start) - start, min(end, other_end) - start):
                    counts[x] += 1
            nbases_overlapping += sum([1 for x in counts if x > 0])

        infile.close()

        return nexons, nexons_overlapping, nbases, nbases_overlapping

    def count(self, filename1, filename2):
        """count overlap between two bed files."""

        E.info("counting started for %s versus %s" % (filename1, filename2))

        idx2 = self.buildIndex(filename2)

        (self.mExons1, self.mExonsOverlapping1,
         self.mBases1, self.mBasesOverlapping1 ) = \
            self._count(filename1, idx2)

        self.mExonsUnique1 = self.mExons1 - self.mExonsOverlapping1
        self.mBasesUnique1 = self.mBases1 - self.mBasesOverlapping1

        idx1 = self.buildIndex(filename1)

        (self.mExons2, self.mExonsOverlapping2,
         self.mBases2, self.mBasesOverlapping2 ) = \
            self._count(filename2, idx1)

        self.mExonsUnique2 = self.mExons2 - self.mExonsOverlapping2
        self.mBasesUnique2 = self.mBases2 - self.mBasesOverlapping2

    def __str__(self):

        return "\t".join(map(str, (
            self.mExons1, self.mExons2,
            self.mExonsOverlapping1, self.mExonsOverlapping2,
            self.mExonsUnique1, self.mExonsUnique2,
            self.mBases1, self.mBases2,
            self.mBasesOverlapping1, self.mBasesOverlapping2,
            self.mBasesUnique1, self.mBasesUnique2 ) ) ) + "\t" +\
            "\t".join([IOTools.prettyPercent(*x) for x in (
                (self.mExonsOverlapping1, self.mExons1),
                (self.mExonsOverlapping2, self.mExons2),
                (self.mExonsUnique1, self.mExons1),
                (self.mExonsUnique2, self.mExons2),
                (self.mBasesOverlapping1, self.mBases1),
                (self.mBasesOverlapping2, self.mBases2),
                (self.mBasesUnique1, self.mBases1),
                (self.mBasesUnique2, self.mBases2))])


class CounterTracks(Counter):

    def __init__(self, filename):
        self.mIndices = Bed.readAndIndex(IOTools.openFile(filename, "r"),
                                         per_track=True)

    def getTracks(self):
        return sorted(self.mIndices.keys())

    def _countIndices(self, idx_in, idx):
        '''count filename against idx.'''

        overlapping_genes = set()
        genes = set()

        # iterate over exons

        nexons, nexons_overlapping = 0, 0
        nbases, nbases_overlapping = 0, 0
        for contig, ix in idx_in.items():

            # note: add a findall function to ncl
            for start, end, value in ix.find(0, 1000000000):
                nexons += 1
                nbases += end - start

                try:
                    intervals = list(idx[contig].find(start, end))
                except KeyError:
                    continue

                if len(intervals) == 0:
                    continue

                nexons_overlapping += 1
                counts = numpy.zeros(end - start, numpy.int)
                for other_start, other_end, other_value in intervals:
                    for x in range(max(start, other_start) - start, min(end, other_end) - start):
                        counts[x] += 1
                nbases_overlapping += sum([1 for x in counts if x > 0])

        return nexons, nexons_overlapping, nbases, nbases_overlapping

    def count(self, filename, track):
        """count overlap between two gtf files."""

        E.info("counting started for %s versus %s" % (filename, track))

        (self.mExons1, self.mExonsOverlapping1,
         self.mBases1, self.mBasesOverlapping1 ) = \
            self._count(filename, self.mIndices[track])

        self.mExonsUnique1 = self.mExons1 - self.mExonsOverlapping1
        self.mBasesUnique1 = self.mBases1 - self.mBasesOverlapping1

        idx = self.buildIndex(filename)

        # count index against index
        (self.mExons2, self.mExonsOverlapping2,
         self.mBases2, self.mBasesOverlapping2 ) = \
            self._countIndices(self.mIndices[track], idx)

        self.mExonsUnique2 = self.mExons2 - self.mExonsOverlapping2
        self.mBasesUnique2 = self.mBases2 - self.mBasesOverlapping2


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $", usage=globals()["__doc__"])

    parser.add_option("-u", "--update", dest="filename_update", type="string",
                      help="if filename is given, previous results will be read from there and only changed sets will be computed [default=%default].")

    parser.add_option("-p", "--pattern-identifier", dest="pattern_id", type="string",
                      help="pattern to convert a filename to an id [default=%default].")

    parser.add_option("-t", "--tracks", dest="tracks", action="store_true",
                      help="compare files against all tracks in the first file [default=%default]")

    parser.set_defaults(
        filename_update=None,
        pattern_id="(.*).bed",
        tracks=None,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if len(args) < 2:
        raise ValueError("at least two arguments required")

    if options.filename_update:
        infile = IOTools.openFile(options.filename_update, "r")
        previous_results = {}
        for line in infile:
            if line.startswith("#"):
                continue
            if line.startswith("set1"):
                continue
            data = line[:-1].split("\t")
            set1, set2 = data[0], data[1]

            if set1 not in previous_results:
                previous_results[set1] = {}
            if set2 not in previous_results:
                previous_results[set2] = {}

            previous_results[set1][set2] = "\t".join(data[2:])
            rev = [(data[x + 1], data[x]) for x in range(2, len(data), 2)]
            previous_results[set2][set1] = "\t".join(IOTools.flatten(rev))
    else:
        previous_results = {}

    pattern_id = re.compile(options.pattern_id)

    def getTitle(x):
        try:
            return pattern_id.search(x).groups()[0]
        except AttributeError:
            return x

    ncomputed, nupdated = 0, 0

    if options.tracks:
        counter = CounterTracks(args[0])
        options.stdout.write("set1\tset2\t%s\n" % counter.getHeader())
        for filename in args[1:]:
            title1 = getTitle(filename)
            for title2 in counter.getTracks():

                if previous_results:
                    try:
                        prev = previous_results[title1][title2]
                    except KeyError:
                        pass
                    else:
                        options.stdout.write(
                            "%s\t%s\t%s\n" % ((title1, title2, prev)))
                        nupdated += 1
                        continue

                counter.count(filename, title2)
                options.stdout.write(
                    "%s\t%s\t%s\n" % ((title1, title2, str(counter))))
                ncomputed += 1
    else:
        counter = Counter()
        options.stdout.write("set1\tset2\t%s\n" % counter.getHeader())

        for x in range(len(args)):

            title1 = getTitle(args[x])

            for y in range(0, x):
                title2 = getTitle(args[y])
                if previous_results:
                    try:
                        prev = previous_results[title1][title2]
                    except KeyError:
                        pass
                    else:
                        options.stdout.write(
                            "%s\t%s\t%s\n" % ((title1, title2, prev)))
                        nupdated += 1
                        continue

                counter.count(args[x], args[y])
                options.stdout.write(
                    "%s\t%s\t%s\n" % ((title1, title2, str(counter))))
                ncomputed += 1

    E.info("nupdated=%i, ncomputed=%i" % (nupdated, ncomputed))
    E.Stop()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""diff_chains.py - compare to chain formatted files
=================================================

:Tags: Genomics GenomeAlignment CHAIN Comparison

Purpose
-------

Compare two genomic alignment files and calculate statistics from the
comparison.

Documentation
--------------

Operates on two `chain
<https://genome.ucsc.edu/goldenPath/help/chain.html>`_ formatted
files.

Outputs a table with the following columns:

+-----------+-------------------------------+
|*Column*   |*Content*                      |
+-----------+-------------------------------+
|contig1    |contig name                    |
+-----------+-------------------------------+
|contig2    |contig name                    |
+-----------+-------------------------------+
|strand     |strand                         |
+-----------+-------------------------------+
|mapped1    |mapped residues                |
+-----------+-------------------------------+
|identical1 |identically mapped residues    |
+-----------+-------------------------------+
|different1 |differently mapped residues    |
+-----------+-------------------------------+
|unique1    |residues mapped only from set1 |
+-----------+-------------------------------+
|pmapped1   |percentage of mapped residues  |
+-----------+-------------------------------+
|pidentical1|percentage of identically      |
|           |mapped residues                |
+-----------+-------------------------------+
|pdifferent1|percentage of differently      |
|           |mapped residues                |
+-----------+-------------------------------+

Similar columns exist for data set 2

Usage
-----

Example::

   cgat diff_chains.py hg19ToMm10v1.chain.over.gz hg19ToMm10v2.chain.over.gz

This will compare the locations that regions within the genome hg19
map to between two different mappings to the genome mm10.

Type::

   python diff_chains.py --help

for command line help.

Command line options
--------------------

"""

import sys
import collections

import CGAT.IOTools as IOTools
import CGAT.Experiment as E
import alignlib_lite


def chain_iterator(infile):
    lines = []
    for line in infile:

        if line.startswith("#"):
            continue
        if line.strip() == "":
            continue
        if line.startswith("chain"):
            if lines:
                yield lines
            lines = []
        lines.append(line)

    yield lines


def validateChain(infile):
    '''validate a chain file.

    No overlapping target coordinates.
    '''

    pairs_t2q = collections.defaultdict(alignlib_lite.py_makeAlignmentBlocks)
    pairs_q2t = collections.defaultdict(alignlib_lite.py_makeAlignmentBlocks)

    for lines in chain_iterator(infile):

        (_,
         _,
         target_contig,
         target_length,
         target_strand,
         target_start,
         target_end,
         query_contig,
         query_length,
         query_strand,
         query_start,
         query_end,
         alignment_id) = lines[0][:-1].split()

        (query_start,
         query_end,
         query_length,
         target_start,
         target_end,
         target_length) = \
            [int(x) for x in
             (query_start,
              query_end,
              query_length,
              target_start,
              target_end,
              target_length)]

        # target_strand is always positive
        assert(target_strand == "+")

        map_target2query = pairs_t2q[target_contig]
        map_query2target = pairs_q2t[query_contig]

        qstart, tstart = query_start, target_start

        for line in lines[1:-1]:
            size, dt, dq = [int(x) for x in line[:-1].split()]
            map_target2query.addDiagonal(tstart,
                                         tstart + size,
                                         0)
            map_query2target.addDiagonal(qstart,
                                         qstart + size,
                                         0)

            qstart += size + dq
            tstart += size + dt

        size = int(lines[-1][:-1])
        map_target2query.addDiagonal(tstart,
                                     tstart + size,
                                     0)
        map_query2target.addDiagonal(qstart,
                                     qstart + size,
                                     0)

    try:
        x = map_query2target.mapRowToCol(0)
    except RuntimeError:
        E.info("query is not unique - this is ok.")

    try:
        x = map_target2query.mapRowToCol(0)
    except RuntimeError:
        E.info("target is not unique")
        return False

    return True


def buildPairs(infile):
    '''read a chain file.

    build target2query alignments.
    The target is always on the positive strand.
    '''
    pairs = collections.defaultdict(alignlib_lite.py_makeAlignmentBlocks)

    def chain_iterator(infile):
        lines = []
        for line in infile:

            if line.startswith("#"):
                continue
            if line.strip() == "":
                continue
            if line.startswith("chain"):
                if lines:
                    yield lines
                lines = []
            lines.append(line)

        yield lines

    for lines in chain_iterator(infile):

        (_,
         _,
         target_contig,
         target_length,
         target_strand,
         target_start,
         target_end,
         query_contig,
         query_length,
         query_strand,
         query_start,
         query_end,
         alignment_id) = lines[0][:-1].split()

        (query_start,
         query_end,
         query_length,
         target_start,
         target_end,
         target_length) = \
            [int(x) for x in
             (query_start,
              query_end,
              query_length,
              target_start,
              target_end,
              target_length)]

        # target_strand is always positive
        assert(target_strand == "+")

        map_target2query = pairs[(target_contig, query_contig, query_strand)]

        qstart, tstart = query_start, target_start

        for line in lines[1:-1]:
            size, dt, dq = [int(x) for x in line[:-1].split()]
            map_target2query.addDiagonal(tstart,
                                         tstart + size,
                                         qstart - tstart)
            qstart += size + dq
            tstart += size + dt

        size = int(lines[-1][:-1])

        map_target2query.addDiagonal(tstart,
                                     tstart + size,
                                     qstart - tstart)

    return pairs

DiffResult = collections.namedtuple(
    "DiffResult", "total same different unique")


def compareChains(pairs1, pairs2):
    '''compare chains in pairs1 versus those in pairs2'''

    result = {}
    for key1, chain1 in pairs1.items():
        E.debug("comparing %s" % str(key1))

        ntotal = chain1.getNumAligned()

        if key1 not in pairs2:
            result[key1] = DiffResult._make((ntotal, 0, 0, ntotal))
            continue

        chain2 = pairs2[key1]
        nsame = alignlib_lite.py_getAlignmentIdentity(
            chain1, chain2, alignlib_lite.py_RR)
        noverlap = alignlib_lite.py_getAlignmentOverlap(
            chain1, chain2, alignlib_lite.py_RR)
        ndifferent = noverlap - nsame
        nunique = ntotal - noverlap

        result[key1] = DiffResult._make((ntotal, nsame, ndifferent, nunique))

    return result


def outputMismatches(pairs1, pairs2,
                     output_mismatches=False,
                     output_unique=False,
                     output_matches=False):
    '''output mismatches.

    This is a very slow operation.
    '''

    outfile = sys.stdout

    for key1, chain1 in pairs1.items():
        E.debug("comparing %s" % str(key1))

        if key1 not in pairs2:
            outfile.write("%s\t%s\t%s\t%i\t%i\t-1\n" %
                          (key1 + (chain1.getRowFrom(), chain2.getRowTo())))
            continue

        chain2 = pairs2[key1]
        reg_start = chain1.getRowFrom()
        for pos in range(chain1.getRowFrom(), chain2.getRowTo()):
            x = chain1.mapRowToCol(pos)
            y = chain2.mapRowToCol(pos)

            if x == y:
                if output_matches:
                    outfile.write("%s\t%s\t%s\t%i\t%i\t%i\t%s\n" %
                                  (key1 + (pos, x, y, "=")))
                continue

            mismatch = x != -1 and y != -1

            if mismatch:
                if output_mismatches:
                    outfile.write("%s\t%s\t%s\t%i\t%i\t%i\t%i\n" %
                                  (key1 + (pos, x, y, x - y)))
            else:
                if output_unique:
                    outfile.write("%s\t%s\t%s\t%i\t%i\t%i\t%s\n" %
                                  (key1 + (pos, x, y, "-")))


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id: chain2psl.py 2899 2010-04-13 14:37:37Z andreas $",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--output-mismatches", dest="output_mismatches", action="store_true",
                      help="output mismatches [%default]")

    parser.add_option("-a", "--output-matches", dest="output_matches", action="store_true",
                      help="output matches [%default]")

    parser.add_option("-u", "--output-unique", dest="output_unique", action="store_true",
                      help="output unique positions [%default]")

    parser.add_option("-r", "--restrict", dest="restrict", type="string",
                      help="restrict analysis to a chromosome pair (chr1:chr1:+) [%default]")

    parser.set_defaults(
        output_mismatches=False,
        output_unique=False,
        restrict=None
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if len(args) != 2:
        raise ValueError("expected two chain files")

    filename_chain1, filename_chain2 = args

    E.info("validating chain 1")
    if not validateChain(IOTools.openFile(filename_chain1)):
        E.warn("validation failed - exiting")
        return 1

    E.info("validating chain 2")
    if not validateChain(IOTools.openFile(filename_chain2)):
        E.warn("validation failed - exiting")
        return 1

    E.info("building pairs for %s" % filename_chain1)
    pairs1 = buildPairs(IOTools.openFile(filename_chain1))
    E.info("read %i pairs" % len(pairs1))

    E.info("building pairs for %s" % filename_chain2)
    pairs2 = buildPairs(IOTools.openFile(filename_chain2))
    E.info("read %i pairs" % len(pairs2))

    if options.restrict:
        restrict = tuple(options.restrict.split(":"))
        pairs1 = {restrict: pairs1[restrict]}
        pairs2 = {restrict: pairs2[restrict]}

    E.info("comparing 1 -> 2")
    comparison1 = compareChains(pairs1, pairs2)
    E.info("comparing 2 -> 1")
    comparison2 = compareChains(pairs2, pairs1)

    all_keys = sorted(list(set(list(comparison1.keys()) + list(comparison2.keys()))))

    outfile = options.stdout
    headers = ("mapped", "identical", "different", "unique")
    outfile.write("contig1\tcontig2\tstrand\t%s\t%s\t%s\t%s\n" %
                  (
                      "\t".join(["%s1" % x for x in headers]),
                      "\t".join(["p%s1" % x for x in headers]),
                      "\t".join(["%s2" % x for x in headers]),
                      "\t".join(["p%s2" % x for x in headers])))

    totals = E.Counter()

    for key in all_keys:
        outfile.write("%s\t%s\t%s" % key)

        if key in comparison1:
            c = comparison1[key]
            outfile.write("\t%i\t%i\t%i\t%i\t" %
                          (c.total, c.same, c.different, c.unique))
            outfile.write(
                "\t".join([IOTools.prettyPercent(x, c.total) for x in c]))

            totals.total1 += c.total
            totals.same1 += c.same
            totals.different1 += c.different
            totals.unique1 += c.unique
        else:
            outfile.write("\t%i\t%i\t%i\t%i\t" % (0, 0, 0, 0))
            outfile.write("\t%i\t%i\t%i\t%i" % (0, 0, 0, 0))

        if key in comparison2:
            c = comparison2[key]
            outfile.write("\t%i\t%i\t%i\t%i\t" %
                          (c.total, c.same, c.different, c.unique))
            outfile.write(
                "\t".join([IOTools.prettyPercent(x, c.total) for x in c]))

            totals.same2 += c.same
            totals.total2 += c.total
            totals.different2 += c.different
            totals.unique2 += c.unique
        else:
            outfile.write("\t%i\t%i\t%i\t%i\t" % (0, 0, 0, 0))
            outfile.write("\t%i\t%i\t%i\t%i" % (0, 0, 0, 0))

        outfile.write("\n")

    outfile.write("total\ttotal\t.\t")
    outfile.write("\t".join(map(str, (totals.total1,
                                      totals.same1,
                                      totals.different1,
                                      totals.unique1,
                                      IOTools.prettyPercent(
                                          totals.total1, totals.total1),
                                      IOTools.prettyPercent(
                                          totals.same1, totals.total1),
                                      IOTools.prettyPercent(
                                          totals.different1, totals.total1),
                                      IOTools.prettyPercent(
                                          totals.unique1, totals.total1),
                                      totals.total2,
                                      totals.same2,
                                      totals.different2,
                                      totals.unique2,
                                      IOTools.prettyPercent(
                                          totals.total2, totals.total2),
                                      IOTools.prettyPercent(
                                          totals.same2, totals.total2),
                                      IOTools.prettyPercent(
                                          totals.different2, totals.total2),
                                      IOTools.prettyPercent(
                                          totals.unique2, totals.total2),
                                      ))) + "\n")

    # output mismapped residues
    if options.output_mismatches or options.output_unique:
        outputMismatches(pairs1, pairs2,
                         output_mismatches=options.output_mismatches,
                         output_unique=options.output_unique,
                         output_matches=options.output_matches,
                         )

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''fastqs2fastqs.py - manipulate (merge/reconcile) fastq files
=============================================================

:Tags: Genomics NGS FASTQ FASTQ Manipulation

Purpose
-------

This script manipulates multiple fastq files and outputs
new fastq files. Currently only the method ``reconcile``
is implemented.

reconcile
+++++++++

Reconcile reads from a pair of fastq files.

This method takes two fastq files and outputs two fastq files such
that all reads in the output are present in both output files.

The typical use case is that two fastq files containing the first and
second part of a read pair have been independently filtered, for
example by quality scores, truncation, etc. As a consequence some
reads might be missing from one file but not the other. The reconcile
method will output two files containing only reads that are common to
both files.

The two files must be sorted by read identifier.

Example input, read2 and read3 are only present in either of the
files:

   # File1        # File 2

   @read1         @read1
   AAA            AAA
   +              +
   !!!            !!!
   @read2         @read3
   CCC            TTT
   +              +
   !!!            !!!
   @read4         @read4
   GGG            GGG
   +              +
   !!!            !!!

Example output, only the reads common to both files are output::

   # File1        # File 2

   @read1         @read1
   AAA            AAA
   +              +
   !!!            !!!
   @read4         @read4
   GGG            GGG
   +              +
   !!!            !!!

Usage
-----

Example::

   python fastqs2fastqs.py \
            --method=reconcile \
            --output-filename-pattern=myReads_reconciled.%s.fastq \
            myReads.1.fastq.gz myReads.2.fastq.gz

In this example we take a pair of fastq files, reconcile by read
identifier and output 2 new fastq files named
``myReads_reconciled.1.fastq.gz`` and
``myReads_reconciled.2.fastq.gz``.

Type::

   python fastqs2fastqs.py --help

for command line help.

Command line options
--------------------

'''

import sys
import re
import CGAT.IOTools as IOTools
import CGAT.Experiment as E


class PatternGetter:

    def __init__(self, pattern):
        self.pattern = re.compile(pattern)

    def __call__(self, id):
        return self.pattern.search(id).groups()[0]


def plain_getter(id):
    return id


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=('reconcile',),
                      help="method to apply [default=%default].")

    parser.add_option(
        "-c", "--chop-identifier", dest="chop", action="store_true",
        help="whether or not to trim last character of the  "
        "sequence name. For example sometimes ids in the first "
        "file in the pair will end with \1 and the second "
        "with \2. If --chop-identifier is not specified "
        "then the results will be wrong [default=%default].")

    parser.add_option(
        "-u", "--unpaired", dest="unpaired", action="store_true",
        help="whether or not to write out unpaired reads "
        "to a separate file")

    parser.add_option(
        "--id-pattern-1", dest="id_pattern_1",
        help="If specified will use the first group from the"
        "pattern to determine the ID for the first read",
        default=None)

    parser.add_option(
        "--id-pattern-2", dest="id_pattern_2",
        help="As above but for read 2",
        default=None)

    parser.add_option(
        "-o", "--output-filename-pattern",
        dest="output_pattern", type="string",
        help="pattern for output files [default=%default].")

    parser.set_defaults(
        method="reconcile",
        chop=False,
        unpaired=False,
        output_pattern="%s.fastq.gz",
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if len(args) != 2:
        raise ValueError(
            "please supply at least two fastq files on the commandline")

    fn1, fn2 = args
    c = E.Counter()

    if options.id_pattern_1:
        id1_getter = PatternGetter(options.id_pattern_1)
    else:
        id1_getter = plain_getter

    if options.id_pattern_2:
        id2_getter = PatternGetter(options.id_pattern_2)
    else:
        id2_getter = plain_getter

    if options.method == "reconcile":

        # IMS: switching to no store second set of read names and only use
        # lazily. Since generators don't have a size must keep track
        id_lengths = {fn1: 0, fn2: 0}

        def getIds(infile, id_getter=plain_getter):
            '''return ids in infile.'''
            aread = infile.readline
            while True:
                l = [aread().rstrip("\r\n") for i in range(4)]
                if not l[0]:
                    break
                r = id_getter(l[0].split()[0])
                # decide if to chop read number off
                id_lengths[infile.name] += 1
                if options.chop:
                    yield r[:-1]
                else:
                    yield r

        def write(outfile, infile, take, unpaired_file=None,
                  id_getter=plain_getter):
            '''filter fastq files with ids in take.'''
            aread = infile.readline
            while True:
                l = [aread().rstrip("\r\n") for i in range(4)]
                if not l[0]:
                    break
                r = id_getter(l[0].split()[0])
                if options.chop:
                    r = r[:-1]
                if r not in take:
                    if unpaired_file is None:
                        continue
                    else:
                        unpaired_file.write("\n".join(l) + "\n")
                else:
                    outfile.write("\n".join(l) + "\n")

        E.info("reading first in pair")
        inf1 = IOTools.openFile(fn1)
        ids1 = set(getIds(inf1, id1_getter))

        E.info("reading second in pair")
        inf2 = IOTools.openFile(fn2)
        # IMS: No longer keep as a set, but lazily evaluate into intersection
        # leads to large memory saving for large inf2, particularly if
        # inf1 is small.
        ids2 = getIds(inf2, id2_getter)
        take = ids1.intersection(ids2)

        E.info("first pair: %i reads, second pair: %i reads, "
               "shared: %i reads" %
               (id_lengths[fn1],
                id_lengths[fn2],
                len(take)))

        if options.unpaired:
            unpaired_filename = IOTools.openFile(
                options.output_pattern % "unpaired", "w")
        else:
            unpaired_filename = None

        with IOTools.openFile(options.output_pattern % "1", "w") as outf:
            inf = IOTools.openFile(fn1)
            E.info("writing first in pair")
            write(outf, inf, take, unpaired_filename, id1_getter)

        with IOTools.openFile(options.output_pattern % "2", "w") as outf:
            inf = IOTools.openFile(fn2)
            E.info("writing second in pair")
            write(outf, inf, take, unpaired_filename, id2_getter)

        if options.unpaired:
            unpaired_filename.close()

    # write footer and output benchmark information.
    E.info("%s" % str(c))
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
fastq2fastq.py - manipulate fastq files
=============================================

:Tags: Genomics NGS Sequences FASTQ Manipulation

Purpose
-------

This script performs manipulations on :term:`fastq` formatted
files. For example it can be used to change the quality score format
or sample a subset of reads.

The script predominantly is used for manipulation of single fastq
files. However, for some of its functionality it will take paired data
using the ``--pair-fastq-file`` and ``--output-filename-pattern`` options.
This applies to the ``sample`` and ``sort`` methods.

Usage
-----

Example::
  In this example we randomly sample 50% of reads from paired data provided in
  two :term:`fastq` files.

   head in.fastq.1

   @SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   =@A@9@BAB@;@BABA?=;@@BB<A@9@;@2>@;??
   @SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCACCCCCCCC
   +SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   =B@9@0>A<B=B=AAA?;*(@A>(@<=*9=9@BA>7
   @SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   AGTGAGCAGGGAAACAATGTCTGTCTAAGAATTTGA

   head in.fastq.2

   +SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   <?@BA?;A=@BA>;@@7###################
   @SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ATTAGTATTATCCATTTATATAATCAATAAAAATGT
   +SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ?ABBA2CCBBB2?=BB@C>=AAC@A=CBB#######
   @SRR111956.5 HWUSI-EAS618:7:1:39:1305 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +SRR111956.5 HWUSI-EAS618:7:1:39:1305 length=36
   AA>5;A>*91?=AAA@@BBA<B=?ABA>2>?A<BB@

   command-line::
     cat in.fastq.1 | python fastq2fastq.py
                      --method=sample --sample-size 0.5
                      --pair-fastq-file in.fastq.2
                      --output-filename-pattern out.fastq.2
                      > out.fastq.1

   head out.fastq.1
   @SRR111956.1 HWUSI-EAS618:7:1:27:1582 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
   +
   =@A@9@BAB@;@BABA?=;@@BB<A@9@;@2>@;??
   @SRR111956.2 HWUSI-EAS618:7:1:29:1664 length=36
   CCCCCCCCCCCCCCCCCCCCCCCCCCCACCCCCCCC
   +
   =B@9@0>A<B=B=AAA?;*(@A>(@<=*9=9@BA>7
   @SRR111956.3 HWUSI-EAS618:7:1:38:878 length=36
   AGTGAGCAGGGAAACAATGTCTGTCTAAGAATTTGA
   +
   <?@BA?;A=@BA>;@@7###################
   @SRR111956.4 HWUSI-EAS618:7:1:38:1783 length=36
   ATTAGTATTATCCATTTATATAATCAATAAAAATGT
   +
   ?ABBA2CCBBB2?=BB@C>=AAC@A=CBB#######

Options
-------

The following methods are implemented (``--method``).

``change-format``

    change the quality format to new format given as
    target-format. Options are ``sanger``,
  ``solexa``, ``phred64``, ``integer`` and ``illumina-1.8``

``sample``

    Sub-sample a fastq file. The size of the sample is set by
    --sample-size

``unique``

    Remove duplicate reads based on read name

``trim3``

    Trim a fixed number of nucleotides from the 3' end of reads.
    (see ``--num-bases``). Note that there are better tools for
   trimming.

``trim5``

    Trim a fixed number of nucleotides from the 5' end of reads.
    (see ``--num-bases``). Note that there are better tools for
   trimming.

``sort``

    Sort the fastq file by read name.

``renumber-reads``

    Rename the reads based on pattern given in ``--pattern-identifier``
    e.g. ``--pattern-identifier="read_%010i"``

Type::

   python fastq2fastq.py --help

for command line help.


Command line options
--------------------

'''

import os
import sys
import re
import random
import CGAT.IOTools as IOTools
import CGAT.Experiment as E
import CGAT.Fastq as Fastq
import CGAT.Genomics as Genomics


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=(
                          "apply",
                          "change-format",
                          "renumber-reads",
                          "sample",
                          "sort",
                          "trim3",
                          "trim5",
                          "unique",
                          "reverse-complement",
                          "grep"),
                      help="method to apply [%default]")

    parser.add_option(
        "--target-format", dest="target_format", type="choice",
        choices=('sanger', 'solexa', 'phred64', 'integer', 'illumina-1.8'),
        help="guess quality score format and set quality scores "
        "to format [default=%default].")

    parser.add_option(
        "--guess-format", dest="guess_format", type="choice",
        choices=('sanger', 'solexa', 'phred64', 'integer', 'illumina-1.8'),
        help="quality score format to assume if ambiguous [default=%default].")

    parser.add_option(
        "--sample-size", dest="sample_size", type="float",
        help="proportion of reads to sample. "
        "Provide a proportion of reads to sample, e.g. 0.1 for 10%, "
        "0.5 for 50%, etc [default=%default].")

    parser.add_option(
        "--pair-fastq-file", dest="pair", type="string",
        help="if data is paired, filename with second pair. "
        "Implemented for sampling [default=%default].")

    parser.add_option(
        "--map-tsv-file", dest="map_tsv_file", type="string",
        help="filename with tab-separated identifiers mapping for "
        "method apply [default=%default].")

    parser.add_option(
        "--num-bases", dest="nbases", type="int",
        help="number of bases to trim [default=%default].")

    parser.add_option(
        "--seed", dest="seed", type="int",
        help="seed for random number generator [default=%default].")

    parser.add_option(
        "--pattern-identifier", dest="renumber_pattern", type="string",
        help="rename reads in file by pattern [default=%default]")

    parser.add_option(
        "--grep-pattern", dest="grep_pattern", type="string",
        help="subset to reads matching pattern [default=%default]")

    parser.set_defaults(
        method=None,
        change_format=None,
        guess_format=None,
        sample_size=0.1,
        nbases=0,
        pair=None,
        apply=None,
        seed=None,
        renumber_pattern="read_%010i",
        grep_pattern=".*")

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    c = E.Counter()

    if options.method is None:
        raise ValueError("no method specified, please use --method")

    if options.method == "change-format":
        for record in Fastq.iterate_convert(options.stdin,
                                            format=options.target_format,
                                            guess=options.guess_format):
            c.input += 1
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "grep":
        for record in Fastq.iterate(options.stdin):
            if re.match(options.grep_pattern, record.seq):
                options.stdout.write("%s\n" % record)

    elif options.method == "reverse-complement":
        for record in Fastq.iterate(options.stdin):
            record.seq = Genomics.complement(record.seq)
            record.quals = record.quals[::-1]
            options.stdout.write("%s\n" % record)

    elif options.method == "sample":
        sample_threshold = min(1.0, options.sample_size)

        random.seed(options.seed)

        if options.pair:
            if not options.output_filename_pattern:
                raise ValueError(
                    "please specify output filename pattern for "
                    "second pair (--output-filename-pattern)")

            outfile1 = options.stdout
            outfile2 = IOTools.openFile(options.output_filename_pattern, "w")

            for record1, record2 in zip(
                    Fastq.iterate(options.stdin),
                    Fastq.iterate(IOTools.openFile(options.pair))):
                c.input += 1
                if random.random() <= sample_threshold:
                    c.output += 1
                    outfile1.write("%s\n" % record1)
                    outfile2.write("%s\n" % record2)
        else:
            for record in Fastq.iterate(options.stdin):
                c.input += 1
                if random.random() <= sample_threshold:
                    c.output += 1
                    options.stdout.write("%s\n" % record)

    elif options.method == "apply":
        ids = set(IOTools.readList(IOTools.openFile(options.apply)))

        for record in Fastq.iterate(options.stdin):
            c.input += 1
            if re.sub(" .*", "", record.identifier).strip() in ids:
                c.output += 1
                options.stdout.write("%s\n" % record)

    elif options.method == "trim3":
        trim3 = options.nbases
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            record.trim(trim3)
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "trim5":
        trim5 = options.nbases
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            record.trim5(trim5)
            options.stdout.write("%s\n" % record)
            c.output += 1

    elif options.method == "unique":
        keys = set()
        for record in Fastq.iterate(options.stdin):
            c.input += 1
            if record.identifier in keys:
                continue
            else:
                keys.add(record.identifier)
            options.stdout.write("%s\n" % record)
            c.output += 1

    # Need to change this to incorporate both pairs
    elif options.method == "sort":
        if not options.pair:
            # This is quicker for a single fastq file
            statement = "paste - - - - | sort -k1,1 -t ' ' | tr '\t' '\n'"
            os.system(statement)
        else:
            if not options.output_filename_pattern:
                raise ValueError(
                    "please specify output filename for second pair "
                    "(--output-filename-pattern)")
            E.warn(
                "consider sorting individual fastq files - "
                "this is memory intensive")
            entries1 = {}
            entries2 = {}

            for record1, record2 in zip(
                    Fastq.iterate(options.stdin),
                    Fastq.iterate(IOTools.openFile(options.pair))):
                entries1[
                    record1.identifier[:-2]] = (record1.seq, record1.quals)
                entries2[
                    record2.identifier[:-2]] = (record2.seq, record2.quals)

            outfile1 = options.stdout
            outfile2 = IOTools.openFile(options.output_filename_pattern, "w")
            assert len(set(entries1.keys()).intersection(
                set(entries2.keys()))) == len(entries1),\
                "paired files do not contain the same reads "\
                "need to reconcile files"

            for entry in sorted(entries1):
                outfile1.write("@%s/1\n%s\n+\n%s\n" %
                               (entry, entries1[entry][0], entries1[entry][1]))
                outfile2.write("@%s/2\n%s\n+\n%s\n" %
                               (entry, entries2[entry][0], entries2[entry][1]))

    elif options.method == "renumber-reads":
        id_count = 1
        for record in Fastq.iterate(options.stdin):
            record.identifier = options.renumber_pattern % id_count
            id_count += 1
            options.stdout.write("@%s\n%s\n+\n%s\n" %
                                 (record.identifier, record.seq, record.quals))

    # write footer and output benchmark information.
    E.info("%s" % str(c))
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Iterators.py - Iterator functions
=================================

A collection of general purpose iterators.

"""

import random


def sample(iterable, sample_size=None):
    """sample # copies from iterator without replacement.

    Stores a temporary copy of the items in iterable. The function has
    thus a possibly high memory footprint and long pre-processing time
    to yield the first element.

    If sample_size is not given, the iterator returns elements in
    random order (see random.shuffle())

    .. note::
        This snippet was downloaded from an unknown source.

    """

    saved = []
    for element in iterable:
        saved.append(element)

    random.shuffle(saved)
    if not sample_size:
        sample_size = len(saved)
    for x in range(sample_size):
        yield saved[x]


def group_by_distance(iterable, distance=1):
    """group integers into non-overlapping intervals that
    are at most *distance* apart.

    >>> list( group_by_distance( (1,1,2,4,5,7) ) )
    [(1, 3), (4, 6), (7, 8)]

    >>> list( group_by_distance( [] ) )
    []

    >>> list( group_by_distance( [3] ) )
    [(3, 4)]

    >>> list( group_by_distance( [3,2] ) )
    Traceback (most recent call last):
    ...
    ValueError: iterable is not sorted: 2 < 3

    .. note:: 
        This snippet was downloaded from an unknown source.

    """
    i = iter(iterable)
    end = None
    start = end = cur = next(i)

    for cur in i:
        if cur < end:
            raise ValueError("iterable is not sorted: %i < %i" % (cur, end))
        if cur - end > distance:
            yield (start, end + 1)
            start = cur
        end = cur
    yield (start, end + 1)

if __name__ == "__main__":
    import doctest
    doctest.testmod()
